# Copyright 2022 StreamSets Inc.

# fmt: off
import uuid

import pytest

from streamsets.sdk.sch_models import PipelineLabel
from streamsets.sdk.utils import SeekableList

# fmt: on


@pytest.mark.snowflake
def test_create_pipeline_with_labels_logic(sch, snowflake_config):
    pipeline_builder = sch.get_pipeline_builder(engine_type='snowflake')

    snowflake_table_name = 'com_streamsets_transformer_snowpark_origin_snowflake_table_TableDOrigin'
    snowflake_table = pipeline_builder.add_stage(name=snowflake_table_name).set_attributes(table='dinos')
    trash = pipeline_builder.add_stage('Trash')
    snowflake_table >> trash
    pipeline_name = 'create_pipeline_with_label_{}'.format(str(uuid.uuid4()))
    label = str(uuid.uuid4())
    pipeline = pipeline_builder.build(title=pipeline_name, labels=[label])
    pipeline.configuration['connectionString'] = snowflake_config['connectionString']
    pipeline.configuration['db'] = snowflake_config['db']
    pipeline.configuration['warehouse'] = snowflake_config['warehouse']
    pipeline.configuration['schema'] = snowflake_config['schema']
    sch.publish_pipeline(pipeline)
    try:
        pipeline_labels = pipeline.labels
        assert len(pipeline_labels) == 1
        assert isinstance(pipeline_labels, SeekableList)
        assert isinstance(pipeline_labels[0], PipelineLabel)
        assert pipeline_labels[0].label == label
    finally:
        sch.delete_pipeline(pipeline)


@pytest.mark.snowflake
def test_update_pipeline_with_labels_logic(sch, snowflake_config):
    pipeline_builder = sch.get_pipeline_builder(engine_type='snowflake')

    snowflake_table_name = 'com_streamsets_transformer_snowpark_origin_snowflake_table_TableDOrigin'
    snowflake_table = pipeline_builder.add_stage(name=snowflake_table_name).set_attributes(table='dinos')
    trash = pipeline_builder.add_stage('Trash')
    snowflake_table >> trash
    pipeline_name = 'update_pipeline_with_label_{}'.format(str(uuid.uuid4()))
    label = str(uuid.uuid4())
    pipeline = pipeline_builder.build(title=pipeline_name)
    pipeline.configuration['connectionString'] = snowflake_config['connectionString']
    pipeline.configuration['db'] = snowflake_config['db']
    pipeline.configuration['warehouse'] = snowflake_config['warehouse']
    pipeline.configuration['schema'] = snowflake_config['schema']
    sch.publish_pipeline(pipeline)
    pipeline.add_label(label)
    sch.publish_pipeline(pipeline)
    try:
        pipeline_labels = pipeline.labels
        assert len(pipeline_labels) == 1
        assert isinstance(pipeline_labels, SeekableList)
        assert isinstance(pipeline_labels[0], PipelineLabel)
        assert pipeline_labels[0].label == label
    finally:
        sch.delete_pipeline(pipeline)


@pytest.mark.snowflake
def test_remove_labels(sch, snowflake_config):
    pipeline_builder = sch.get_pipeline_builder(engine_type='snowflake')

    snowflake_table_name = 'com_streamsets_transformer_snowpark_origin_snowflake_table_TableDOrigin'
    snowflake_table = pipeline_builder.add_stage(name=snowflake_table_name).set_attributes(table='dinos')
    trash = pipeline_builder.add_stage('Trash')
    snowflake_table >> trash
    pipeline_name = 'create_pipeline_with_label_{}'.format(str(uuid.uuid4()))
    label1, label2 = str(uuid.uuid4()), str(uuid.uuid4())
    pipeline = pipeline_builder.build(title=pipeline_name, labels=[label1, label2])
    pipeline.configuration['connectionString'] = snowflake_config['connectionString']
    pipeline.configuration['db'] = snowflake_config['db']
    pipeline.configuration['warehouse'] = snowflake_config['warehouse']
    pipeline.configuration['schema'] = snowflake_config['schema']
    sch.publish_pipeline(pipeline)
    try:
        pipeline_labels = pipeline.labels
        assert len(pipeline_labels) == 2
        assert isinstance(pipeline_labels, SeekableList)
        assert isinstance(pipeline_labels[0], PipelineLabel)

        pipeline.remove_label(label1)
        sch.publish_pipeline(pipeline)

        pipeline_labels = pipeline.labels
        assert len(pipeline_labels) == 1
        assert isinstance(pipeline_labels, SeekableList)
        assert isinstance(pipeline_labels[0], PipelineLabel)
        assert pipeline_labels[0].label == label2
    finally:
        sch.delete_pipeline(pipeline)



# Copyright 2022 StreamSets Inc.

# fmt: off
import pytest

# fmt: on


def test_update_user(sch, test_user, test_group):
    """Test updating a user's group membership and roles."""
    roles_to_update = ['Engine Administrator', 'Job Operator', 'Pipeline Editor', 'Deployment Manager']
    test_user.roles = roles_to_update
    all_group = sch.groups.get(display_name='all')
    test_user.groups = [all_group, test_group]
    sch.update_user(test_user)

    updated_user = sch.users.get(email_address=test_user.email_address)
    for role in roles_to_update:
        assert role in updated_user.roles
    assert len(updated_user.roles) == len(roles_to_update)
    assert updated_user.groups.get(group_id=test_group.group_id)
    assert updated_user.groups.get(group_id=all_group.group_id)


def test_activate_and_deactivate_user(sch, test_user):
    """Test activation and deactivation of a user."""
    try:
        # Grabbing user again to make sure we have the latest activation status, rather than making two API calls below
        user = sch.users.get(email_address=test_user.email_address)
        if user.active:
            sch.deactivate_user(user)
            assert not sch.users.get(email_address=user.email_address).active
            sch.activate_user(user)
            assert sch.users.get(email_address=user.email_address).active
        else:
            sch.activate_user(user)
            assert sch.users.get(email_address=user.email_address).active
            sch.deactivate_user(user)
            assert not sch.users.get(email_address=user.email_address).active
    finally:
        # Put the user back into an 'ACTIVE' state as expected at the start of tests
        if not sch.users.get(email_address=test_user.email_address).active:
            sch.activate_user(test_user)


def test_user_org_admin_role(sch, test_user):
    """Test assigning and removing the Org Admin role for a user."""
    sch.assign_administrator_role(test_user)
    updated_test_user = sch.users.get(email_address=test_user.email_address)
    assert 'Organization Administrator' in updated_test_user.roles
    assert 'org-admin' in updated_test_user._data['roles']

    sch.remove_administrator_role(test_user)
    updated_test_user = sch.users.get(email_address=test_user.email_address)
    assert 'Organization Administrator' not in updated_test_user.roles
    assert 'org-user' in updated_test_user._data['roles']


def test_update_group(sch, test_user, test_group):
    """Test updating a group's assigned roles and membership."""
    try:
        role_to_add = 'Topology User'
        assert test_user not in test_group.users
        assert role_to_add not in test_group.roles
        test_group.users.append(test_user)
        test_group.roles.append(role_to_add)
        sch.update_group(test_group)
        # update_group updates the in-memory representation directly
        assert test_group.users.get(email_address=test_user.email_address)
        assert role_to_add in test_group.roles
    finally:
        # Put the group back to its original state
        # Retrieve the test user again since its definition was updated via role/group addition
        test_group.users.remove(test_user)
        test_group.roles.remove(role_to_add)
        sch.update_group(test_group)
        # We expect this to throw a ValueError since the user was removed from the group
        with pytest.raises(ValueError) as value_error:
            test_group.users.get(email_address=test_user.email_address)
        assert value_error
        assert role_to_add not in test_group.roles



# Copyright 2024 StreamSets Inc.

# fmt: off
import pytest

from streamsets.sdk.exceptions import EnginelessError
from streamsets.sdk.utils import get_random_string

# fmt: on


def test_default_engineless_pipeline(sch):
    pipeline_builder = sch.get_pipeline_builder()
    dev_raw_data_source = pipeline_builder.add_stage('Dev Raw Data Source')
    trash = pipeline_builder.add_stage('Trash')
    dev_raw_data_source >> trash
    pipeline = pipeline_builder.build("Engineless_Pipeline_{}".format(get_random_string()))
    with pytest.raises(EnginelessError):
        sch.publish_pipeline(pipeline)
    sch.delete_pipeline(pipeline)


def test_versioned_engineless_pipeline(sch):
    pipeline_builder = sch.get_pipeline_builder(engine_version_id="DC:5.6.2::Released")
    dev_raw_data_source = pipeline_builder.add_stage('Dev Raw Data Source')
    trash = pipeline_builder.add_stage('Trash')
    dev_raw_data_source >> trash
    pipeline = pipeline_builder.build("Engineless_Pipeline_{}".format(get_random_string()))
    with pytest.raises(EnginelessError):
        sch.publish_pipeline(pipeline)
    sch.delete_pipeline(pipeline)



#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2024

# fmt: off
import pytest

from streamsets.sdk.analytics import (
    FUNCTION_CALLS_HEADERS_KEY, INTERACTIVE_MODE_HEADERS_KEY, MAX_THRESHOLD, AnalyticsHeaders,
)

# fmt: on
DUMMY_VALUE = 'dummy value'


@pytest.fixture(scope="function")
def analytics_headers():
    headers = {FUNCTION_CALLS_HEADERS_KEY: None, INTERACTIVE_MODE_HEADERS_KEY: 'False'}
    return headers


def test_analytics_attribute_added_to_headers(sch_sdc_pipeline_builder, analytics_headers):
    pipeline_builder = sch_sdc_pipeline_builder
    analytics_instance = AnalyticsHeaders.get_instance()

    pipeline_builder.dummy_attribute1 = DUMMY_VALUE
    analytics_headers[FUNCTION_CALLS_HEADERS_KEY] = 'PipelineBuilder.dummy_attribute1'
    assert analytics_instance.headers == analytics_headers

    # FUNCTION_CALLS_HEADERS_KEY is flushed after the function has been called.
    analytics_headers[FUNCTION_CALLS_HEADERS_KEY] = None

    # Set another attribute to assert that dummy_attribute1 was flushed out
    pipeline_builder.dummy_attribute2 = DUMMY_VALUE
    analytics_headers[FUNCTION_CALLS_HEADERS_KEY] = 'PipelineBuilder.dummy_attribute2'
    assert analytics_instance.headers == analytics_headers


def test_analytics_method_added_to_headers(sch_sdc_pipeline_builder, analytics_headers):
    pipeline_builder = sch_sdc_pipeline_builder
    analytics_instance = AnalyticsHeaders.get_instance()

    pipeline_builder.add_stage('Snowflake', type='destination')
    analytics_headers[FUNCTION_CALLS_HEADERS_KEY] = 'PipelineBuilder.add_stage'
    assert analytics_instance.headers == analytics_headers


def test_analytics_property_added_to_headers(sch_sdc_pipeline, analytics_headers):
    pipeline = sch_sdc_pipeline
    analytics_instance = AnalyticsHeaders.get_instance()

    pipeline.name = DUMMY_VALUE
    analytics_headers[FUNCTION_CALLS_HEADERS_KEY] = 'Pipeline.name'
    assert analytics_instance.headers == analytics_headers

    # FUNCTION_CALLS_HEADERS_KEY is flushed after the function has been called.
    analytics_headers[FUNCTION_CALLS_HEADERS_KEY] = None

    pipeline.name
    analytics_headers[FUNCTION_CALLS_HEADERS_KEY] = 'Pipeline.name'
    assert analytics_instance.headers == analytics_headers


def test_analytics_resets_headers_if_called_more_than_threshold(sch_sdc_pipeline, analytics_headers):
    pipeline = sch_sdc_pipeline
    analytics_instance = AnalyticsHeaders.get_instance()

    for i in range(0, MAX_THRESHOLD + 1):
        pipeline.name

    analytics_headers[FUNCTION_CALLS_HEADERS_KEY] = 'Pipeline.name'
    assert analytics_instance.headers == analytics_headers



# fmt: off
from copy import deepcopy

import pytest

from streamsets.sdk.sch_api import Command
from streamsets.sdk.sch_models import CollectionModelResults, Engine, Engines

from .resources.engines_data import GET_ENGINE_RESPONSE_COLLECTOR_JSON, GET_ENGINE_RESPONSE_TRANSFORMER_JSON

# fmt: on

NUM_OF_ENGINES = 5


class MockControlHub:
    def __init__(self, value):
        self.api_client = MockApiClient(value)
        self.organization = 'DUMMY ORG'


class MockApiClient:
    def __init__(self, value):
        self.value = value

    def get_engine(self, **kwargs):
        return Command(self, MockResponse(self.value))

    def get_all_registered_engines(self, **kwargs):
        return Command(self, MockResponse(self.value))


class MockResponse:
    def __init__(self, value):
        self.value = value

    def json(self):
        return self.value


@pytest.fixture(scope="function")
def get_engine_response_collector_json():
    json = deepcopy(GET_ENGINE_RESPONSE_COLLECTOR_JSON)

    return json


@pytest.fixture(scope="function")
def get_engine_response_transformer_json():
    json = deepcopy(GET_ENGINE_RESPONSE_TRANSFORMER_JSON)

    return json


@pytest.fixture(scope="function")
def get_all_registered_engines_response_collector_json(get_engine_response_collector_json):
    json = {'totalCount': 1, 'offset': 50, 'len': 50, 'data': []}
    get_engine_response_collector_json['id'] += '{}'
    for i in range(NUM_OF_ENGINES):
        get_engine_response_collector_json['id'].format(i)
        json['data'].append(get_engine_response_collector_json)

    return json


@pytest.fixture(scope="function")
def get_all_registered_engines_response_transformer_json(get_engine_response_transformer_json):
    json = {'totalCount': 1, 'offset': 50, 'len': 50, 'data': []}
    get_engine_response_transformer_json['id'] += '{}'
    for i in range(NUM_OF_ENGINES):
        get_engine_response_transformer_json['id'].format(i)
        json['data'].append(get_engine_response_transformer_json)

    return json


@pytest.mark.parametrize(
    "engine_json",
    ["get_all_registered_engines_response_collector_json", "get_all_registered_engines_response_transformer_json"],
)
def test_engines_sanity(engine_json, request):
    engine_json = request.getfixturevalue(engine_json)

    mock_control_hub = MockControlHub(engine_json)
    e = Engines(mock_control_hub)

    response = e._get_all_results_from_api()

    assert isinstance(response, CollectionModelResults)
    assert response.results == engine_json
    assert len(response.results['data']) == NUM_OF_ENGINES
    assert response.class_type == Engine
    assert response.class_kwargs['control_hub'] == mock_control_hub


@pytest.mark.parametrize(
    "engine_json",
    ["get_engine_response_collector_json", "get_engine_response_transformer_json"],
)
def test_engines_get_with_id(engine_json, request):
    engine_json = request.getfixturevalue(engine_json)
    mock_control_hub = MockControlHub(engine_json)
    e = Engines(mock_control_hub)

    response = e._get_all_results_from_api(id='123')

    assert isinstance(response, CollectionModelResults)
    assert response.results == [engine_json]
    assert response.class_type == Engine
    assert response.class_kwargs['control_hub'] == mock_control_hub


@pytest.mark.parametrize(
    "engine_json,engine_type",
    [("get_engine_response_collector_json", "COLLECTOR"), ("get_engine_response_transformer_json", "TRANSFORMER")],
)
def test_engines_get_with_id_and_type(engine_json, engine_type, request):
    engine_json = request.getfixturevalue(engine_json)
    mock_control_hub = MockControlHub(engine_json)
    e = Engines(mock_control_hub)

    response = e._get_all_results_from_api(engine_type=engine_type, id='123')

    assert isinstance(response, CollectionModelResults)
    assert response.results == [engine_json]
    assert response.class_type == Engine
    assert response.class_kwargs['control_hub'] == mock_control_hub


@pytest.mark.parametrize(
    "engine_json,engine_type",
    [("get_engine_response_collector_json", "TRANSFORMER"), ("get_engine_response_transformer_json", "COLLECTOR")],
)
def test_engines_get_with_incorrect_type(engine_json, engine_type, request):
    engine_json = request.getfixturevalue(engine_json)
    e = Engines(MockControlHub(engine_json))

    with pytest.raises(TypeError):
        e._get_all_results_from_api(engine_type=engine_type, id='123')


def test_engines_get_with_invalid_type(get_engine_response_collector_json):
    e = Engines(MockControlHub(get_engine_response_collector_json))

    with pytest.raises(ValueError):
        e._get_all_results_from_api(engine_type='DUMMY TYPE', id='123')



# Copyright 2024 StreamSets Inc.

# fmt: off
import pytest
import requests

from streamsets.sdk.sch_models import (
    FinishCondition, Job, JobSequence, JobSequenceBuilder, JobSequenceHistoryLog, JobSequenceHistoryLogs, Step,
)
from streamsets.sdk.utils import SeekableList, get_random_string

# fmt: on
START_TIME_2099 = 4200268800
END_TIME_2099 = 4200854400
START_TIME_2100 = 4237401600
END_TIME_2100 = 4237996800
UTC_TIME_Z0NE = 'UTC'
BASIC_CRON_TAB_MASK = '0/1 * 1/1 * ? *'


@pytest.fixture(scope="module")
def simple_pipeline(sch, sch_authoring_sdc_id):
    """A trivial pipeline:

    dev_data_generator >> trash
    """
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    trash = pipeline_builder.add_stage('Trash')
    dev_data_generator >> trash
    pipeline = pipeline_builder.build('simple_pipeline_{}'.format(get_random_string()))
    sch.publish_pipeline(pipeline)

    try:
        yield pipeline
    finally:
        sch.api_client.delete_pipeline(pipeline.pipeline_id)


@pytest.fixture(scope="module")
def sample_jobs(sch, simple_pipeline, sch_executor_sdc_label):
    """A set of simple jobs based on simple pipeline."""
    job_builder = sch.get_job_builder()

    jobs = []
    for i in range(5):
        job = job_builder.build('test_simple_job_fetch_{}'.format(i), pipeline=simple_pipeline)
        job.data_collector_labels = sch_executor_sdc_label
        sch.add_job(job)
        jobs.append(job)

    try:
        yield jobs
    finally:
        for job in jobs:
            sch.delete_job(job)


def test_publish_job_sequence(sch):
    job_sequence_builder = sch.get_job_sequence_builder()
    assert isinstance(job_sequence_builder, JobSequenceBuilder)

    job_sequence_builder.add_start_condition(START_TIME_2099, END_TIME_2099, UTC_TIME_Z0NE, BASIC_CRON_TAB_MASK)

    job_sequence = job_sequence_builder.build(
        name='Sequence {}'.format(get_random_string()), description='description {}'.format(get_random_string())
    )
    assert isinstance(job_sequence, JobSequence)
    assert not job_sequence.id
    try:
        sch.publish_job_sequence(job_sequence=job_sequence)
        assert job_sequence.id
        assert job_sequence.start_time == START_TIME_2099
        assert job_sequence.end_time == END_TIME_2099
        assert job_sequence.timezone == UTC_TIME_Z0NE
        assert job_sequence.crontab_mask == BASIC_CRON_TAB_MASK

        job_sequence_from_sch = sch.job_sequences.get(id=job_sequence.id)
        assert job_sequence_from_sch.id == job_sequence.id
        assert job_sequence_from_sch.name == job_sequence.name
        assert job_sequence_from_sch.description == job_sequence.description
    except Exception as e:
        raise e
    finally:
        if job_sequence.id:
            sch.api_client.delete_job_sequence(job_sequence.id)


def test_delete_job_sequence(sch):
    job_sequence_builder = sch.get_job_sequence_builder()
    job_sequence_builder.add_start_condition(START_TIME_2099, END_TIME_2099, UTC_TIME_Z0NE, BASIC_CRON_TAB_MASK)
    job_sequence = job_sequence_builder.build(
        name='Sequence {}'.format(get_random_string()), description='description {}'.format(get_random_string())
    )
    try:
        deleted = False
        sch.publish_job_sequence(job_sequence=job_sequence)
        job_sequence_from_sch = sch.job_sequences.get(id=job_sequence.id)
        sch.delete_job_sequences(job_sequence_from_sch)

        with pytest.raises(requests.exceptions.HTTPError):
            sch.job_sequences.get(id=job_sequence_from_sch.id)

        deleted = True
    except Exception as e:
        raise e
    finally:
        if job_sequence.id and not deleted:
            sch.api_client.delete_job_sequence(job_sequence.id)


def test_update_job_sequence_metadata(sch):
    job_sequence_builder = sch.get_job_sequence_builder()
    job_sequence_builder.add_start_condition(START_TIME_2099, END_TIME_2099, UTC_TIME_Z0NE, BASIC_CRON_TAB_MASK)
    job_sequence = job_sequence_builder.build(
        name='Sequence {}'.format(get_random_string()), description='description {}'.format(get_random_string())
    )
    try:
        sch.publish_job_sequence(job_sequence=job_sequence)
        assert job_sequence.start_time == START_TIME_2099
        assert job_sequence.end_time == END_TIME_2099
        assert job_sequence.timezone == UTC_TIME_Z0NE
        assert job_sequence.crontab_mask == BASIC_CRON_TAB_MASK

        job_sequence = sch.job_sequences.get(id=job_sequence.id)
        job_sequence.start_time = START_TIME_2100
        job_sequence.end_time = END_TIME_2100
        new_desc = get_random_string()
        job_sequence.description = new_desc
        assert job_sequence.start_time == START_TIME_2100
        assert job_sequence.end_time == END_TIME_2100
        assert job_sequence.description == new_desc

        sch.update_job_sequence_metadata(job_sequence=job_sequence)

        job_sequence = sch.job_sequences.get(id=job_sequence.id)
        assert job_sequence.start_time == START_TIME_2100
        assert job_sequence.end_time == END_TIME_2100
        assert job_sequence.description == new_desc

    except Exception as e:
        raise e
    finally:
        sch.api_client.delete_job_sequence(job_sequence.id)


def test_job_sequences_property(sch):
    try:
        job_sequence_list = []
        number_of_job_sequences = 3
        for i in range(number_of_job_sequences):
            job_sequence_builder = sch.get_job_sequence_builder()
            job_sequence_builder.add_start_condition(START_TIME_2099, END_TIME_2099, UTC_TIME_Z0NE, BASIC_CRON_TAB_MASK)
            job_sequence = job_sequence_builder.build(
                name='TEST Sequence {}'.format(get_random_string()),
                description='description {}'.format(get_random_string()),
            )
            sch.publish_job_sequence(job_sequence=job_sequence)
            job_sequence_list.append(job_sequence.id)
        job_sequences_from_sch = sch.job_sequences

        counter = 0
        for job_sequence in job_sequences_from_sch:
            if job_sequence.id in job_sequence_list:
                counter += 1

        assert counter == number_of_job_sequences

    except Exception as e:
        raise e
    finally:
        if job_sequence_list:
            sch.api_client.delete_job_sequences(job_sequence_list)


def test_get_and_delete_history_log_and_history_logs(sch, sample_jobs):
    job_sequence_builder = sch.get_job_sequence_builder()
    job_sequence_builder.add_start_condition(START_TIME_2099, END_TIME_2099, UTC_TIME_Z0NE, BASIC_CRON_TAB_MASK)
    job_sequence = job_sequence_builder.build(
        name='Sequence {}'.format(get_random_string()), description='description {}'.format(get_random_string())
    )
    try:
        sch.publish_job_sequence(job_sequence=job_sequence)
        job_sequence = sch.job_sequences.get(id=job_sequence.id)

        job1 = sample_jobs[0]
        job_sequence.add_step_with_jobs([job1])
        sch.enable_job_sequence(job_sequence)

        # Start and Stop the job sequence
        job_sequence.refresh()
        sch.run_job_sequence(job_sequence)
        sch.wait_for_job_sequence_status(job_sequence, 'ACTIVE')
        sch.wait_for_job_status(job1, 'ACTIVE')
        sch.stop_job(job1)
        job1.refresh()
        sch.wait_for_job_sequence_status(job_sequence, 'INACTIVE')
        job_sequence.refresh()

        logs = job_sequence.get_history_log()
        assert isinstance(logs, SeekableList)

        logs = job_sequence.history_logs
        assert isinstance(logs, JobSequenceHistoryLogs)
        assert isinstance(logs[0], JobSequenceHistoryLog)

        logs = job_sequence.history_logs.get_all()
        assert isinstance(logs, SeekableList)
        assert isinstance(logs[0], JobSequenceHistoryLog)

        with pytest.raises(TypeError):
            job_sequence.delete_history_logs(['dummy_value'])

        assert len(logs) != 0
        job_sequence.delete_history_logs(logs)
        assert len(job_sequence.history_logs) == 0

    except Exception as e:
        raise e
    finally:
        if job_sequence.id:
            sch.api_client.delete_job_sequence(job_sequence.id)


def test_add_steps_and_jobs_to_job_sequence(sch, sample_jobs):
    job_sequence_builder = sch.get_job_sequence_builder()

    job_sequence_builder.add_start_condition(START_TIME_2099, END_TIME_2099, UTC_TIME_Z0NE, BASIC_CRON_TAB_MASK)

    job_sequence = job_sequence_builder.build(
        name='Sequence {}'.format(get_random_string()), description='description {}'.format(get_random_string())
    )
    job1, job2 = sample_jobs[0], sample_jobs[1]
    try:
        sch.publish_job_sequence(job_sequence=job_sequence)

        # Add step and pull from SCH
        job_sequence.add_step_with_jobs([job1])
        job_sequence_from_sch = sch.job_sequences.get(id=job_sequence.id)

        # Make sure step has been added
        assert len(job_sequence_from_sch.steps) == 1
        step = job_sequence_from_sch.steps[0]
        assert isinstance(step, Step)

        # Add job to step and pull again
        step.add_jobs([job2])
        job_sequence_again_from_sch = sch.job_sequences.get(id=job_sequence.id)

        # Make sure extra job has been added to step and only one step exists within the Job Sequence
        assert len(job_sequence_again_from_sch.steps) == 1
        step = job_sequence_again_from_sch.steps[0]
        assert len(step.step_jobs) == 2
        assert isinstance(step.step_jobs[0], Job) and isinstance(step.step_jobs[1], Job)
        job_ids = [job.job_id for job in step.step_jobs]
        assert job1.job_id in job_ids and job2.job_id in job_ids

        # test changing step names
        job_sequence_again_from_sch = sch.job_sequences.get(id=job_sequence.id)
        step = job_sequence_again_from_sch.steps[0]
        assert step.name == 'Step 1'
        step.name = 'My new step'

        job_sequence_again_from_sch = sch.job_sequences.get(id=job_sequence.id)
        step = job_sequence_again_from_sch.steps[0]
        assert step.name == 'My new step'

    except Exception as e:
        raise e
    finally:
        if job_sequence.id:
            sch.api_client.delete_job_sequence(job_sequence.id)


def test_reorder_and_remove_jobs_from_sequence(sch, sample_jobs):
    job_sequence_builder = sch.get_job_sequence_builder()
    job_sequence_builder.add_start_condition(START_TIME_2099, END_TIME_2099, UTC_TIME_Z0NE, BASIC_CRON_TAB_MASK)
    job_sequence = job_sequence_builder.build(
        name='Sequence {}'.format(get_random_string()), description='description {}'.format(get_random_string())
    )
    try:
        sch.publish_job_sequence(job_sequence=job_sequence)
        job_sequence.add_step_with_jobs(sample_jobs)

        job_sequence_from_sch = sch.job_sequences.get(id=job_sequence.id)
        assert len(job_sequence_from_sch.steps) == 5

        steps = job_sequence_from_sch.steps

        # base case: make sure that our step_1_id & step_3_id are in index 0,2
        step_1_id, step_3_id = steps[0].id, steps[2].id
        assert steps[0].id == step_1_id
        assert steps[0].step_number == 1
        assert steps[2].id == step_3_id
        assert steps[2].step_number == 3

        # trigger move_step & assert it has changed in SCH
        job_sequence_from_sch.move_step(steps[0], steps[2].step_number, swap=True)
        job_sequence_from_sch = sch.job_sequences.get(id=job_sequence.id)
        new_steps = job_sequence_from_sch.steps
        assert new_steps[0].id == step_3_id
        assert new_steps[2].id == step_1_id

        # Remove and assert it has change in local memory and assert it's changed in SCH
        job_sequence_from_sch.remove_step(job_sequence_from_sch.steps[2])
        assert len(job_sequence_from_sch.steps) == 4
        job_sequence_from_sch = sch.job_sequences.get(id=job_sequence.id)
        assert len(job_sequence_from_sch.steps) == 4

    except Exception as e:
        raise e
    finally:
        sch.api_client.delete_job_sequence(job_sequence.id)


def test_change_status_of_job_sequence(sch, sample_jobs):
    job_sequence_builder = sch.get_job_sequence_builder()
    job_sequence_builder.add_start_condition(START_TIME_2099, END_TIME_2099, UTC_TIME_Z0NE, BASIC_CRON_TAB_MASK)
    job_sequence = job_sequence_builder.build(
        name='Sequence {}'.format(get_random_string()), description='description {}'.format(get_random_string())
    )
    job1 = sample_jobs[0]

    try:
        sch.publish_job_sequence(job_sequence=job_sequence)
        job_sequence.add_step_with_jobs([job1])

        job_sequence = sch.job_sequences.get(id=job_sequence.id)
        assert job_sequence.status == 'DISABLED'

        sch.enable_job_sequence(job_sequence)
        job_sequence = sch.job_sequences.get(id=job_sequence.id)
        assert job_sequence.status == 'INACTIVE'

        sch.run_job_sequence(job_sequence)
        job_sequence = sch.job_sequences.get(id=job_sequence.id)
        assert job_sequence.status == 'ACTIVE'

        sch.stop_job(job1)
        job_sequence = sch.job_sequences.get(id=job_sequence.id)
        assert job_sequence.status == 'INACTIVE'

    except Exception as e:
        raise e
    finally:
        if job_sequence.id:
            sch.api_client.delete_job_sequence(job_sequence.id)


def test_job_sequence_finish_conditions(sch, sample_jobs):
    # Set up Job Sequence and Jobs
    job_sequence_builder = sch.get_job_sequence_builder()
    job_sequence_builder.add_start_condition(START_TIME_2099, END_TIME_2099, UTC_TIME_Z0NE, BASIC_CRON_TAB_MASK)
    job_sequence = job_sequence_builder.build(
        name='Sequence {}'.format(get_random_string()), description='description {}'.format(get_random_string())
    )
    job = sample_jobs[0]
    try:
        # Publish and create relevant Step with Job
        sch.publish_job_sequence(job_sequence=job_sequence)
        job_sequence.add_step_with_jobs([job])
        job_sequence_from_sch = sch.job_sequences.get(id=job_sequence.id)
        step = job_sequence_from_sch.steps[0]

        # Assert finish condition has been properly created
        finish_condition = step.create_finish_condition(
            condition_type='CRON', job=job, crontab_mask=BASIC_CRON_TAB_MASK
        )
        assert isinstance(finish_condition, FinishCondition)
        assert step.finish_conditions[0].id == finish_condition.id

        # Assert changes have been made in Platform
        job_sequence_from_sch = sch.job_sequences.get(id=job_sequence.id)
        step = job_sequence_from_sch.steps[0]
        assert step.finish_conditions[0].id == finish_condition.id

        # Make in-memory changes to Finish Condition
        finish_condition = step.finish_conditions[0]
        finish_condition.condition_type = 'END_TIME'
        finish_condition.end_time = END_TIME_2100
        finish_condition.timezone = UTC_TIME_Z0NE

        # Assert changes reflect in-memory
        assert finish_condition.condition_type == 'END_TIME'
        assert finish_condition.end_time == END_TIME_2100
        assert finish_condition.timezone == UTC_TIME_Z0NE

        # Reflect in Platform
        step.update_finish_condition(finish_condition)

        # Assert changes have been made in Platform
        job_sequence_from_sch = sch.job_sequences.get(id=job_sequence.id)
        step = job_sequence_from_sch.steps[0]
        finish_condition = step.finish_conditions[0]
        assert finish_condition.condition_type == 'END_TIME'
        assert finish_condition.end_time == END_TIME_2100
        assert finish_condition.timezone == UTC_TIME_Z0NE

        # Make sure the appropriate job is returned
        returned_job = finish_condition.job
        assert returned_job.job_id == job.job_id

        # Delete Finish Condition
        step.delete_finish_condition(finish_condition)
        assert len(step.finish_conditions) == 0

        # Assert changes have been made in Platform
        job_sequence_from_sch = sch.job_sequences.get(id=job_sequence.id)
        step = job_sequence_from_sch.steps[0]
        assert len(step.finish_conditions) == 0

    except Exception as e:
        raise e

    finally:
        # Cleanup Job Sequence
        if job_sequence.id:
            sch.api_client.delete_job_sequence(job_sequence.id)


def test_running_a_job_sequence(sch, sample_jobs):
    job_sequence_builder = sch.get_job_sequence_builder()
    job_sequence_builder.add_start_condition(START_TIME_2099, END_TIME_2099, UTC_TIME_Z0NE, BASIC_CRON_TAB_MASK)
    job_sequence = job_sequence_builder.build(
        name='Sequence {}'.format(get_random_string()), description='description {}'.format(get_random_string())
    )

    try:
        sch.publish_job_sequence(job_sequence=job_sequence)

        job_sequence.add_step_with_jobs(sample_jobs)
        step = job_sequence.steps[2]
        assert all(step.status != 'RUNNING' for step in job_sequence.steps)
        sch.enable_job_sequence(job_sequence)

        job_sequence = sch.job_sequences.get(id=job_sequence.id)
        sch.run_job_sequence(job_sequence, start_from_step_number=step.step_number, single_step=True)

        job_sequence = sch.job_sequences.get(id=job_sequence.id)
        step = job_sequence.steps[2]
        assert step.status == 'RUNNING'
        assert all(step.status != 'RUNNING' for step in job_sequence.steps if step.step_number != 3)
        sch.stop_job(step.step_jobs[0])
        job_sequence = sch.job_sequences.get(id=job_sequence.id)

        sch.run_job_sequence(job_sequence, start_from_step_number=4, single_step=False)
        job_sequence = sch.job_sequences.get(id=job_sequence.id)
        step = job_sequence.steps[3]
        assert step.status == 'RUNNING'
        assert all(step.status != 'RUNNING' for step in job_sequence.steps if step.step_number != 4)
        sch.stop_job(step.step_jobs[0])

        step = job_sequence.steps[4]
        sch.wait_for_job_status(step.step_jobs[0], 'ACTIVE')

        job_sequence = sch.job_sequences.get(id=job_sequence.id)
        step = job_sequence.steps[4]
        assert step.status == 'RUNNING'
        assert all(step.status != 'RUNNING' for step in job_sequence.steps if step.step_number != 5)
        sch.stop_job(step.step_jobs[0])

    except Exception as e:
        raise e

    finally:
        if job_sequence.id:
            job_sequence = sch.job_sequences.get(id=job_sequence.id)
            if job_sequence.status == 'ACTIVE':
                sch.disable_job_sequence(job_sequence)

            sch.api_client.delete_job_sequence(job_sequence.id)



#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2024

# fmt: off
from copy import deepcopy

import pytest

from streamsets.sdk.sch_models import DeploymentBuilder, Pipeline
from streamsets.sdk.sch_models import PipelineBuilder as SchSdcPipelineBuilder
from streamsets.sdk.sch_models import StPipelineBuilder as SchStPipelineBuilder
from streamsets.sdk.sdc_models import PipelineBuilder as SdcPipelineBuilder
from streamsets.sdk.st_models import PipelineBuilder as StPipelineBuilder

from .resources.conftest_data import (
    AZURE_ENVIRONMENT_JSON, DEPLOYMENT_BUILDER_JSON, PIPELINE_BUILDER_DEFINITIONS_JSON, PIPELINE_BUILDER_JSON,
)

# fmt: on


@pytest.fixture(scope="function")
def pipeline_builder_json():
    data = deepcopy(PIPELINE_BUILDER_JSON)
    return data


@pytest.fixture(scope="function")
def pipeline_builder_definitions():
    data = deepcopy(PIPELINE_BUILDER_DEFINITIONS_JSON)
    return data


class DummyEngines:
    def get(self, id):
        return DummyEngine()


class DummyEngine:
    def __init__(self):
        self._data = {"startUpTime": 0}
        self._sch_pipeline = {"sdcId": 0}


class MockControlHub:
    def __init__(self):
        self.organization = "12345"

    @property
    def engines(self):
        return DummyEngines()


@pytest.fixture(scope="function")
def sdc_pipeline_builder(pipeline_builder_json, pipeline_builder_definitions):
    return SdcPipelineBuilder(pipeline=pipeline_builder_json, definitions=pipeline_builder_definitions)


@pytest.fixture(scope="function")
def sch_sdc_pipeline_builder(pipeline_builder_json, sdc_pipeline_builder):
    return SchSdcPipelineBuilder(
        pipeline=pipeline_builder_json,
        data_collector_pipeline_builder=sdc_pipeline_builder,
        control_hub=MockControlHub(),
    )


@pytest.fixture(scope="function")
def st_pipeline_builder(pipeline_builder_json, pipeline_builder_definitions):
    return StPipelineBuilder(pipeline=pipeline_builder_json, definitions=pipeline_builder_definitions)


@pytest.fixture(scope="function")
def sch_st_pipeline_builder(pipeline_builder_json, st_pipeline_builder):
    return SchStPipelineBuilder(
        pipeline=pipeline_builder_json,
        transformer_pipeline_builder=st_pipeline_builder,
        control_hub=MockControlHub(),
    )


@pytest.fixture(scope="function")
def deployment_builder_json():
    data = deepcopy(DEPLOYMENT_BUILDER_JSON)
    return data


@pytest.fixture(scope="function")
def self_deployment_builder(deployment_builder_json):
    additional_attributes = {"type": "SELF", "installType": None}

    data = {**deployment_builder_json, **additional_attributes}

    return DeploymentBuilder(data, MockControlHub())


@pytest.fixture(scope="function")
def gce_deployment_builder(deployment_builder_json):
    additional_attributes = {
        "type": "GCE",
        "initScript": None,
        "machineType": None,
        "region": None,
        "allowedResourceLocations": None,
        "secretManagerReplicationPolicy": None,
        "zones": None,
        "blockProjectSshKeys": None,
        "publicSshKey": None,
        "trackingUrl": None,
        "instanceServiceAccountEmail": None,
        "tags": None,
        "resourceLabels": None,
        "subnetwork": None,
        "attachPublicIp": None,
    }

    data = {**deployment_builder_json, **additional_attributes}

    return DeploymentBuilder(data, MockControlHub())


@pytest.fixture(scope="function")
def ec2_deployment_builder(deployment_builder_json):
    additional_attributes = {
        "type": "EC2",
        "initScript": None,
        "instanceType": None,
        "resourceTags": None,
        "sshKeySource": None,
        "sshKeyPairName": None,
        "instanceProfileArn": None,
        "trackingUrl": None,
    }

    data = {**deployment_builder_json, **additional_attributes}

    return DeploymentBuilder(data, MockControlHub())


@pytest.fixture(scope="function")
def azure_deployment_builder(deployment_builder_json):
    additional_attributes = {
        "type": "AZURE_VM",
        "initScript": None,
        "vmSize": None,
        "sshKeySource": None,
        "sshKeyPairName": None,
        "publicSshKey": None,
        "resourceGroup": None,
        "resourceTags": None,
        "managedIdentity": None,
        "trackingUrl": None,
        "zones": None,
        "attachPublicIp": None,
    }

    data = {**deployment_builder_json, **additional_attributes}

    return DeploymentBuilder(data, MockControlHub())


@pytest.fixture(scope="function")
def kubernetes_deployment_builder(deployment_builder_json):
    additional_attributes = {
        "type": "KUBERNETES",
        "kubernetesLabels": None,
        "memoryRequest": None,
        "cpuRequest": None,
        "memoryLimit": None,
        "cpuLimit": None,
        "yaml": None,
        "advancedMode": None,
        "hpa": None,
        "hpaMinReplicas": None,
        "hpaMaxReplicas": None,
        "hpaTargetCPUUtilizationPercentage": None,
    }

    data = {**deployment_builder_json, **additional_attributes}

    return DeploymentBuilder(data, MockControlHub())


@pytest.fixture(scope="function")
def pipeline_definitions():
    return {
        "stages": [
            {  # Snowflake destination
                "instanceName": "Snowflake_01",
                "library": "streamsets-datacollector-sdc-snowflake-lib",
                "stageName": "com_streamsets_pipeline_stage_destination_snowflake_SnowflakeDTarget",
                "stageVersion": "18",
                "configuration": [],
                "services": [],
                "uiInfo": {
                    "yPos": 50,
                    "stageType": "TARGET",
                    "description": "",
                    "label": "Snowflake 1",
                    "xPos": 60,
                },
                "inputLanes": [],
                "outputLanes": [],
                "eventLanes": [],
            }
        ]
    }


@pytest.fixture(scope="function")
def pipeline_data_json(pipeline_definitions, request):
    return {
        "pipelineId": 1,
        "commitId": None,
        "name": "Test Pipeline",
        "version": 1,
        "executorType": "COLLECTOR" if "sdc" in request.node.name else "TRANSFORMER",
        "pipelineDefinitions": pipeline_definitions,
    }


@pytest.fixture(scope="function")
def sch_sdc_pipeline(
    pipeline_data_json,
    pipeline_definitions,
    sch_sdc_pipeline_builder,
    mocker,
):
    return Pipeline(
        pipeline=pipeline_data_json,
        pipeline_definition=pipeline_definitions,
        rules_definition=None,
        library_definitions={},
        control_hub=mocker.Mock(),
        builder=sch_sdc_pipeline_builder,
    )


@pytest.fixture(scope="function")
def sch_st_pipeline(
    pipeline_data_json,
    pipeline_definitions,
    sch_st_pipeline_builder,
    mocker,
):
    return Pipeline(
        pipeline=pipeline_data_json,
        pipeline_definition=pipeline_definitions,
        rules_definition=None,
        library_definitions={},
        control_hub=mocker.Mock(),
        builder=sch_st_pipeline_builder,
    )


@pytest.fixture(scope="function")
def azure_environment_json():
    data = deepcopy(AZURE_ENVIRONMENT_JSON)
    return data



# Copyright 2023 StreamSets Inc.

# fmt: off
import pytest

from streamsets.sdk.sch_models import SAQLSearch, SAQLSearchBuilder

# fmt: on


@pytest.fixture(scope="function")
def saql_search_json():
    saql_search = {"id": None, "creator": None, "name": None, "createTime": None, "query": ''}
    return saql_search


def set_type_and_mode(saql_search_type, saql_mode_type, saql_search_json):
    saql_search_json['type'] = saql_search_type
    saql_search_json['mode'] = saql_mode_type


@pytest.mark.parametrize(
    'properties',
    (
        {'name': 1, 'operator': 'contains', 'value': '', 'condition': 'AND'},
        {'name': 'name', 'operator': 1, 'value': '', 'condition': 'AND'},
        {'name': 'name', 'operator': 'contains', 'value': 1, 'condition': 'AND'},
        {'name': 'name', 'operator': 'contains', 'value': '', 'condition': 1},
    ),
)
def test_add_filter_invalid_parameter_type(mocker, properties):
    mock_control_hub = mocker.Mock()
    saql_builder = SAQLSearchBuilder(mock_control_hub, SAQLSearch.PipelineType.PIPELINE, SAQLSearch.ModeType.BASIC)

    # Invalid parameter types raise TypeError
    with pytest.raises(TypeError):
        saql_builder.add_filter(
            property_name=properties['name'],
            property_operator=properties['operator'],
            property_value=properties['value'],
            property_condition_combiner=properties['condition'],
        )


def test_add_filter_invalid_mode_type(mocker, saql_search_json):
    mock_control_hub = mocker.Mock()
    set_type_and_mode(SAQLSearch.PipelineType.PIPELINE.value, SAQLSearch.ModeType.ADVANCED.value, saql_search_json)
    saql_builder = SAQLSearchBuilder(saql_search_json, mock_control_hub, SAQLSearch.ModeType.ADVANCED.value)

    assert not saql_builder.query

    # Invalid mode raises ValueError
    with pytest.raises(ValueError):
        saql_builder.add_filter()


def test_add_filter_invalid_property_operator(mocker, saql_search_json):
    mock_control_hub = mocker.Mock()
    set_type_and_mode(SAQLSearch.PipelineType.PIPELINE.value, SAQLSearch.ModeType.BASIC.value, saql_search_json)
    saql_builder = SAQLSearchBuilder(saql_search_json, mock_control_hub, SAQLSearch.ModeType.BASIC.value)

    # Invalid property operator raises ValueError
    with pytest.raises(ValueError):
        saql_builder.add_filter(property_operator='foo')


def test_add_filter_invalid_property_name(mocker, saql_search_json):
    mock_control_hub = mocker.Mock()
    set_type_and_mode(SAQLSearch.PipelineType.PIPELINE.value, SAQLSearch.ModeType.BASIC.value, saql_search_json)
    saql_builder = SAQLSearchBuilder(saql_search_json, mock_control_hub, SAQLSearch.ModeType.BASIC.value)

    # Invalid property name raises ValueError
    with pytest.raises(ValueError):
        saql_builder.add_filter(property_name='foo')


@pytest.mark.parametrize(
    'invalid_name_to_type',
    (
        {'name': 'failover', 'type': SAQLSearch.PipelineType.PIPELINE},
        {'name': 'failover', 'type': SAQLSearch.PipelineType.FRAGMENT},
        {'name': 'label', 'type': SAQLSearch.JobType.JOB_TEMPLATE},
        {'name': 'draft', 'type': SAQLSearch.JobType.JOB_INSTANCE},
        {'name': 'engine_label', 'type': SAQLSearch.JobType.JOB_DRAFT_RUN},
    ),
)
def test_add_filter_invalid_property_name_for_saql_type(mocker, saql_search_json, invalid_name_to_type):
    mock_control_hub = mocker.Mock()
    set_type_and_mode(invalid_name_to_type['type'].value, SAQLSearch.ModeType.BASIC.value, saql_search_json)
    saql_builder = SAQLSearchBuilder(saql_search_json, mock_control_hub, SAQLSearch.ModeType.BASIC.value)

    # Invalid operator to saql type raises ValueError
    with pytest.raises(ValueError):
        saql_builder.add_filter(property_name=invalid_name_to_type['name'])


@pytest.mark.parametrize(
    "saql_search_type, property_name, property_operator,property_value, property_condition_combiner",
    [
        (SAQLSearch.PipelineType.PIPELINE, 'name', 'contains', '', 'AND'),
        (SAQLSearch.PipelineType.FRAGMENT, 'name', 'contains', '', 'AND'),
        (SAQLSearch.JobType.JOB_TEMPLATE, 'name', 'contains', '', 'AND'),
        (SAQLSearch.JobType.JOB_INSTANCE, 'name', 'contains', '', 'AND'),
        (SAQLSearch.JobType.JOB_DRAFT_RUN, 'name', 'contains', '', 'AND'),
        (SAQLSearch.JobType.JOB_RUN, 'status', '==', '', 'AND'),
        (SAQLSearch.JobType.JOB_SEQUENCE, 'name', '==', '', 'AND'),
    ],
)
def test_add_filter_default_values(
    mocker,
    saql_search_json,
    saql_search_type,
    property_name,
    property_operator,
    property_value,
    property_condition_combiner,
):
    mock_control_hub = mocker.Mock()
    set_type_and_mode(saql_search_type.value, SAQLSearch.ModeType.BASIC.value, saql_search_json)
    saql_builder = SAQLSearchBuilder(saql_search_json, mock_control_hub, SAQLSearch.ModeType.BASIC.value)

    saql_builder.add_filter(
        property_name=property_name,
        property_operator=property_operator,
        property_value=property_value,
        property_condition_combiner=property_condition_combiner,
    )

    # Only one query should exist
    assert saql_builder.query[0].get('filter')['name'] == property_name
    assert saql_builder.query[0].get('operator') == property_operator
    assert saql_builder.query[0].get('value') == property_value
    assert saql_builder.query[0].get('conditionCombiner') == property_condition_combiner


def test_build_invalid_name_type(mocker, saql_search_json):
    mock_control_hub = mocker.Mock()
    set_type_and_mode(SAQLSearch.PipelineType.PIPELINE.value, SAQLSearch.ModeType.BASIC.value, saql_search_json)
    saql_builder = SAQLSearchBuilder(saql_search_json, mock_control_hub, SAQLSearch.ModeType.BASIC.value)

    # Invalid parameter types raise TypeError
    with pytest.raises(TypeError):
        saql_builder.build(name=1)


def test_build_empty_filter_basic_mode(mocker, saql_search_json):
    mock_control_hub = mocker.Mock()
    set_type_and_mode(SAQLSearch.PipelineType.PIPELINE.value, SAQLSearch.ModeType.BASIC.value, saql_search_json)
    saql_builder = SAQLSearchBuilder(saql_search_json, mock_control_hub, SAQLSearch.ModeType.BASIC.value)

    # Default values when running empty build() in basic mode
    property_name, property_operator = 'name', 'contains'
    property_value, property_condition_combiner = '', 'AND'
    saql_search_obj = saql_builder.build('foo')

    # Only one query should exist
    assert saql_builder.query[0].get('filter')['name'] == property_name
    assert saql_builder.query[0].get('operator') == property_operator
    assert saql_builder.query[0].get('value') == property_value
    assert saql_builder.query[0].get('conditionCombiner') == property_condition_combiner
    assert isinstance(saql_search_obj, SAQLSearch)  # Sanity check


def test_build_empty_filter_advanced_mode(mocker, saql_search_json):
    mock_control_hub = mocker.Mock()
    set_type_and_mode(SAQLSearch.PipelineType.PIPELINE.value, SAQLSearch.ModeType.ADVANCED.value, saql_search_json)
    saql_builder = SAQLSearchBuilder(saql_search_json, mock_control_hub, SAQLSearch.ModeType.ADVANCED.value)

    saql_search_obj = saql_builder.build('foo')
    assert isinstance(saql_search_obj, SAQLSearch)  # Sanity check



# Copyright 2022 StreamSets Inc.

# fmt: off
import json
import os
import re
import time
from datetime import datetime

import pytest

from streamsets.sdk.sch_models import ACL, Configuration, Pipeline
from streamsets.sdk.utils import SeekableList, get_random_string

# fmt: on

NUM_PIPELINES = 12
PIPELINE_NAME_PREFIX = '{}_pipeline_sdc_test'.format(str(datetime.now()))


@pytest.fixture(scope="module")
def sample_pipelines(resources_label, sch, sch_authoring_sdc_id):
    """A set of trivial pipelines:

    dev_data_generator >> trash
    """
    pipelines = []

    for i in range(NUM_PIPELINES):
        pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

        dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
        trash = pipeline_builder.add_stage('Trash')
        dev_data_generator >> trash

        pipeline_name = PIPELINE_NAME_PREFIX + '_{}_{}'.format(i, get_random_string())
        pipeline = pipeline_builder.build(pipeline_name)
        sch.publish_pipeline(pipeline)

        pipelines.append(pipeline)

    try:
        yield pipelines
    finally:
        for pipeline in pipelines:
            sch.delete_pipeline(pipeline)


@pytest.fixture(scope="module")
def sample_pipeline(resources_label, sch, sch_authoring_sdc_id):
    """A trivial pipelines:

    dev_data_generator >> trash
    """
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    trash = pipeline_builder.add_stage('Trash')
    dev_data_generator >> trash
    pipeline = pipeline_builder.build(
        'pipeline_sdc_test_{}_{}_{}'.format(resources_label, NUM_PIPELINES + 1, get_random_string())
    )
    sch.publish_pipeline(pipeline)

    try:
        yield pipeline
    finally:
        sch.delete_pipeline(pipeline)


def test_iter(sch, sample_pipelines):
    iter_results = []
    for pipeline in sch.pipelines:
        iter_results.append(pipeline.pipeline_id)
    assert len(iter_results) >= len(sample_pipelines)
    assert not ({pipeline.pipeline_id for pipeline in sample_pipelines} - set(iter_results))


def test_pipelines_get_returns_pipeline_object(sch, sample_pipelines):
    assert sample_pipelines[0] in sch.pipelines
    assert isinstance(sch.pipelines.get(organization=sch.organization), Pipeline)


def test_pipeline_in_returns_true(sch, sample_pipelines):
    assert sample_pipelines[0] in sch.pipelines


def test_pipeline_contains(sch, sample_pipelines):
    pipeline = sample_pipelines[0]
    assert sch.pipelines.contains(pipeline_id=pipeline.pipeline_id)
    assert sch.pipelines.contains(name=pipeline.name)
    assert not sch.pipelines.contains(pipeline_id='impossible_to_clash_with_this_zbsdudcugeqwgui')


def test_pipelines_len_works(sch, sample_pipelines):
    assert len(sch.pipelines) >= len(sample_pipelines)


# Tests for __getitem__
def test_pipelines_getitem_int_pos_invalid_index(sch):
    total_number_of_pipelines = len(sch.pipelines)
    with pytest.raises(IndexError):
        sch.pipelines[total_number_of_pipelines]


def test_pipelines_getitem_int_neg_invalid_index(sch):
    total_number_of_pipelines = len(sch.pipelines)
    with pytest.raises(IndexError):
        sch.pipelines[-total_number_of_pipelines - 1]


def test_pipelines_getitem_int_instance_type(sch, sample_pipelines):
    total_number_of_pipelines = len(sch.pipelines)
    assert isinstance(sch.pipelines[total_number_of_pipelines - 1], Pipeline)


@pytest.mark.skip("This test is flakey and will be skipped until TLKT-1052 is resolved")
def test_pipelines_getitem_slice_all(sch):
    total_number_of_pipelines = len(sch.pipelines)
    assert len(sch.pipelines[:]) == total_number_of_pipelines


@pytest.mark.skip("This test is flakey and will be skipped until TLKT-1052 is resolved")
def test_pipelines_getitem_slice_stop(sch):
    assert len(sch.pipelines[:NUM_PIPELINES]) == NUM_PIPELINES


@pytest.mark.skip("This test is flakey and will be skipped until TLKT-1052 is resolved")
def test_pipelines_getitem_slice_start_stop(sch):
    assert len(sch.pipelines[0:NUM_PIPELINES]) == NUM_PIPELINES


@pytest.mark.skip("This test is flakey and will be skipped until TLKT-1052 is resolved")
def test_pipelines_getitem_slice_neg_start(sch):
    total_number_of_pipelines = len(sch.pipelines)
    assert len(sch.pipelines[-total_number_of_pipelines:]) == total_number_of_pipelines


@pytest.mark.skip("This test is flakey and will be skipped until TLKT-1052 is resolved")
def test_pipelines_getitem_slice_neg_start_0_stop(sch):
    total_number_of_pipelines = len(sch.pipelines)
    # stop is 0 so, it should always return empty list
    assert len(sch.pipelines[-total_number_of_pipelines:0]) == 0


@pytest.mark.skip("This test is flakey and will be skipped until TLKT-1052 is resolved")
def test_pipelines_getitem_slice_extra_neg_start(sch):
    total_number_of_pipelines = len(sch.pipelines)
    # eg. [1,2,3,4][-9:] = [1,2,3,4]
    assert len(sch.pipelines[-total_number_of_pipelines - 5 :]) == total_number_of_pipelines


@pytest.mark.skip("This test is flakey and will be skipped until TLKT-1052 is resolved")
def test_pipelines_getitem_slice_step_size(sch):
    total_number_of_pipelines = len(sch.pipelines)
    # verify with step size for both the cases of total_number_of_pipelines being an odd or even number
    assert len(sch.pipelines[::2]) == total_number_of_pipelines // 2 + total_number_of_pipelines % 2


@pytest.mark.skip("This test is flakey and will be skipped until TLKT-1052 is resolved")
def test_pipelines_getitem_slice_neg_step_size(sch):
    # verify negative step size
    assert len(sch.pipelines[-1:-2:-1]) == 1


# Tests for get_all
def test_get_all_returns_seekable_list(sch, sample_pipeline):
    assert isinstance(sch.pipelines.get_all(), SeekableList)


def test_get_all_filter_by_organization_and_version(sch, sample_pipeline):
    assert sch.pipelines.get_all(organization=sch.organization, version='1')


def test_get_all_filter_by_text(sch, sample_pipelines):
    pipelines = sch.pipelines.get_all(filter_text=PIPELINE_NAME_PREFIX)
    assert len(pipelines) == NUM_PIPELINES


# ACL tests
def test_pipeline_acl_get(sch):
    pipeline = sch.pipelines[0]
    acl = pipeline.acl
    assert isinstance(acl, ACL)
    assert acl.resource_type == 'PIPELINE'


def test_import_export_pipelines(sch, sch_authoring_sdc_id):
    # Create a pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    trash = pipeline_builder.add_stage('Trash')
    dev_data_generator >> trash
    pipeline = pipeline_builder.build('import_export_pipeline')
    sch.publish_pipeline(pipeline)

    pipelines_zip_data = sch.export_pipelines([pipeline])
    assert isinstance(pipelines_zip_data, bytes)

    path = '/tmp'
    filename = 'import_export_pipelines_{}.zip'.format(datetime.now().strftime('%Y_%m_%d_%H_%M_%S'))
    with open('{}/{}'.format(path, filename), 'wb') as pipelines_zip_file:
        pipelines_zip_file.write(pipelines_zip_data)
    pipelines_zip_file_name = '{}/{}'.format(path, filename)

    try:
        pipelines_before = [pipeline]
        sch.delete_pipeline(pipeline)
        with open(pipelines_zip_file_name, 'rb') as input_file:
            pipelines_after = sch.import_pipelines_from_archive(input_file, 'Test Import Pipelines')
        assert len(pipelines_before) == len(pipelines_after)
    except Exception as ex:
        raise ex
    finally:
        os.remove(pipelines_zip_file_name)
        if pipeline:
            sch.delete_pipeline(pipeline)


def test_import_pipelines_v2(sch, sch_authoring_sdc_id):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    trash = pipeline_builder.add_stage('Trash')
    dev_data_generator >> trash
    pipeline = pipeline_builder.build('import_pipelines_v2')
    try:
        sch.publish_pipeline(pipeline)

        pipeline_id = pipeline.pipeline_id
        assert len(sch.pipelines.get_all(pipeline_id=pipeline_id)) == 1

        pipelines_zip_data = sch.export_pipelines([pipeline])
        sch.delete_pipeline(pipeline)
        assert len(sch.pipelines.get_all(pipeline_id=pipeline_id)) == 0

        pipelines = sch.api_client.verify_import_pipelines(
            "Test Import Pipelines V2", pipelines_zip_data
        ).response.json()
        sch.api_client.import_pipelines_v2(pipelines)

        assert len(sch.pipelines.get_all(pipeline_id=pipeline_id)) == 1
    except Exception as ex:
        raise ex
    finally:
        if pipeline.pipeline_id:
            sch.delete_pipeline(pipeline)


def test_verify_import_pipelines(sch, sch_authoring_sdc_id):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    trash = pipeline_builder.add_stage('Trash')
    dev_data_generator >> trash
    pipeline = pipeline_builder.build('verify_import_pipeline')
    sch.publish_pipeline(pipeline)

    pipelines_zip_data = sch.export_pipelines([pipeline])
    sch.delete_pipeline(pipeline)
    import_pipeline_json = sch.api_client.verify_import_pipelines(
        'Test Verify Import Pipelines',
        pipelines_zip_data,
    ).response.json()
    import_pipeline_json_keys = ['commitPipelineJsons', 'importConnectionJsons', 'availableConnections']
    for key in import_pipeline_json_keys:
        assert key in import_pipeline_json.keys()


def test_get_pipeline_templates(sch):
    templates = sch.pipelines.get_all(template=True)
    assert isinstance(templates, SeekableList)
    if templates:
        assert isinstance(templates[0], Pipeline)


def test_update_pipeline(sch, sch_authoring_sdc_id):
    # Build and publish pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    trash = pipeline_builder.add_stage('Trash')
    dev_data_generator >> trash
    pipeline = pipeline_builder.build('pipeline_to_be_updated')
    sch.publish_pipeline(pipeline)

    try:
        # Update batch size
        for stage in pipeline.stages:
            if stage.stage_name == 'com_streamsets_pipeline_stage_devtest_RandomDataGeneratorSource':
                assert stage.batch_size != 999
                stage.batch_size = 999
        sch.publish_pipeline(pipeline)

        # Confirm pipeline.stages is updated
        for stage in pipeline.stages:
            if stage.stage_name == 'com_streamsets_pipeline_stage_devtest_RandomDataGeneratorSource':
                assert stage.batch_size == 999

        # Confirm it is updated in json data
        for stage in json.loads(pipeline._data['pipelineDefinition'])['stages']:
            if stage['stageName'] == 'com_streamsets_pipeline_stage_devtest_RandomDataGeneratorSource':
                for config in stage['configuration']:
                    if config['name'] == 'batchSize':
                        assert config['value'] == 999
    finally:
        sch.delete_pipeline(pipeline)


def test_update_pipeline_with_transformer(sch, sch_authoring_transformer_id):
    # Build and publish pipeline
    if len(sch.transformers) < 1:
        pytest.skip("Need at least one transformer instance to run this test")
    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)
    dev_random = pipeline_builder.add_stage('Dev Random')
    trash = pipeline_builder.add_stage('Trash')
    dev_random >> trash
    pipeline = pipeline_builder.build('pipeline_to_be_updated')
    sch.publish_pipeline(pipeline)

    try:
        # Update batch size
        for stage in pipeline.stages:
            if stage.stage_name == 'com_streamsets_pipeline_spark_origin_dev_DevRandomDOrigin':
                assert stage.batch_size != 999
                stage.batch_size = 999
        sch.publish_pipeline(pipeline)

        # Confirm pipeline.stages is updated
        for stage in pipeline.stages:
            if stage.stage_name == 'com_streamsets_pipeline_spark_origin_dev_DevRandomDOrigin':
                assert stage.batch_size == 999

        # Confirm it is updated in json data
        for stage in json.loads(pipeline._data['pipelineDefinition'])['stages']:
            if stage['stageName'] == 'com_streamsets_pipeline_spark_origin_dev_DevRandomDOrigin':
                for config in stage['configuration']:
                    if config['name'] == 'batchSize':
                        assert config['value'] == 999
    finally:
        sch.delete_pipeline(pipeline)


@pytest.mark.parametrize('executor_type', ('TRANSFORMER', 'DATACOLLECTOR'))
def test_execution_mode_config(sch, executor_type, sch_authoring_sdc_id, sch_authoring_transformer_id):
    if executor_type == 'TRANSFORMER':
        if len(sch.transformers) < 1:
            pytest.skip("Need at least one transformer instance to run this test")
        pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)
        dev = pipeline_builder.add_stage('Dev Random')
    elif executor_type == 'DATACOLLECTOR':
        if len(sch.data_collectors) < 1:
            pytest.skip("Need at least one registered DataCollector instance to run this test")
        pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
        dev = pipeline_builder.add_stage('Dev Data Generator')

    trash = pipeline_builder.add_stage('Trash')
    dev >> trash
    pipeline = pipeline_builder.build('execution_mode_config')
    sch.publish_pipeline(pipeline)

    try:
        pipeline_definition = json.loads(pipeline._data['pipelineDefinition'])
        assert 'executionMode' not in pipeline_definition
        configuration = Configuration(configuration=pipeline_definition['configuration'])
        assert configuration.get('executionMode', None) is not None
    finally:
        sch.delete_pipeline(pipeline)


def test_duplicate_pipeline(sch, sample_pipeline):
    duplicated_pipelines = sch.duplicate_pipeline(sample_pipeline)
    try:
        assert isinstance(duplicated_pipelines, SeekableList)
        assert len(duplicated_pipelines) == 1
        assert (
            sch.pipelines.get(name='{} copy'.format(sample_pipeline.name), only_published=False).pipeline_id
            == duplicated_pipelines[0].pipeline_id
        )
    finally:
        for pipeline in duplicated_pipelines:
            sch.delete_pipeline(pipeline)


def test_duplicate_pipelines(sch, sample_pipeline):
    name = 'test_duplicate_pipeline_{}_'.format(get_random_string())
    duplicated_pipelines = sch.duplicate_pipeline(
        sample_pipeline, name=name, description='Testing Duplication', number_of_copies=3
    )

    pipeline_id_to_object = {}
    for pipeline in duplicated_pipelines:
        pipeline_id_to_object[pipeline.pipeline_id] = pipeline

    try:
        assert isinstance(duplicated_pipelines, SeekableList)
        assert len(duplicated_pipelines) == 3

        for duplicate_pipeline in duplicated_pipelines:
            pipeline = sch.pipelines.get(name=duplicate_pipeline.name, only_published=False)

            assert pipeline.pipeline_id in pipeline_id_to_object
            assert pipeline.name == pipeline_id_to_object[pipeline.pipeline_id].name

    finally:
        for pipeline in duplicated_pipelines:
            sch.delete_pipeline(pipeline)


def test_create_and_publish_draft(sch, sch_authoring_sdc_id):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    trash = pipeline_builder.add_stage('Trash')
    dev_data_generator >> trash
    pipeline = pipeline_builder.build('simple pipeline draft')
    sch.publish_pipeline(pipeline, draft=True)
    try:
        assert pipeline.draft is True
        assert pipeline.name == 'simple pipeline draft'
        assert pipeline.version == '1-DRAFT'
        sch.publish_pipeline(pipeline)
        assert getattr(pipeline, 'draft', False) is False
        assert pipeline.name == 'simple pipeline draft'
    finally:
        sch.delete_pipeline(pipeline)


def test_publishing_duplicate_of_pipeline_draft(sch, sch_authoring_sdc_id):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    trash = pipeline_builder.add_stage('Trash')
    dev_data_generator >> trash
    pipeline = pipeline_builder.build('simple pipeline draft')
    sch.publish_pipeline(pipeline, draft=True)
    try:
        duplicated_pipeline = sch.duplicate_pipeline(pipeline)[0]
        assert duplicated_pipeline.draft is True
        assert duplicated_pipeline.version == '1-DRAFT'
        duplicated_pipeline_name = duplicated_pipeline.name
        sch.publish_pipeline(duplicated_pipeline)
        assert getattr(duplicated_pipeline, 'draft', False) is False
        assert duplicated_pipeline.name == duplicated_pipeline_name
    finally:
        sch.delete_pipeline(duplicated_pipeline)
        sch.delete_pipeline(pipeline)


def test_should_fail_to_add_empty_collection_of_downstream_stages_to_collector(sch, sch_authoring_sdc_id):
    if len(sch.data_collectors) < 1:
        pytest.skip("Need at least one registered DataCollector instance to run this test")

    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')

    with pytest.raises(ValueError) as exc_info:
        dev_data_generator >> []

    assert exc_info.match(re.escape("Attempted to add an empty collection of downstream stages."))


def test_should_fail_to_add_empty_collection_of_downstream_stages_to_spark_transformer(
    sch, sch_authoring_transformer_id
):
    if len(sch.transformers) < 1:
        pytest.skip("Need at least one transformer instance to run this test")

    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)
    dev_random = pipeline_builder.add_stage('Dev Random')

    with pytest.raises(ValueError) as exc_info:
        dev_random >> []

    assert exc_info.match(re.escape("Attempted to add an empty collection of downstream stages."))


def test_should_fail_to_add_empty_collection_of_downstream_stages_to_snowflake_transformer(sch):
    pipeline_builder = sch.get_pipeline_builder(engine_type='snowflake', fragment=True)
    dev_data_generator = pipeline_builder.add_stage('Snowflake Table', type='origin')

    with pytest.raises(ValueError) as exc_info:
        dev_data_generator >> []

    assert exc_info.match(re.escape("Attempted to add an empty collection of downstream stages."))


def test_retrieve_draft_pipeline_via_pipeline_id(sch, sch_authoring_sdc_id):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    trash = pipeline_builder.add_stage('Trash')
    dev_data_generator >> trash
    pipeline = pipeline_builder.build('simple pipeline draft')
    sch.publish_pipeline(pipeline, draft=True)
    try:
        draft_pipeline = sch.pipelines.get(pipeline_id=pipeline.pipeline_id, draft=True)
        assert pipeline.draft is True
        assert draft_pipeline.draft is True
        assert draft_pipeline.pipeline_id == pipeline.pipeline_id
        assert draft_pipeline.commit_id == pipeline.commit_id
    finally:
        sch.delete_pipeline(pipeline)


def test_pipeline_add_stage(sch, sch_authoring_sdc_id):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    trash = pipeline_builder.add_stage('Trash')
    dev_data_generator >> trash
    sample_pipeline = pipeline_builder.build('sample_pipeline_{}'.format(get_random_string()))
    sch.publish_pipeline(sample_pipeline)

    try:
        # Assert that the inital stages are present
        assert len(sample_pipeline.stages) == 2
        sample_pipeline.stages[0].instance_name = 'DevDataGenerator_01'
        sample_pipeline.stages[1].instance_name = 'Trash_01'

        # Assert that the stage has been added in memory
        sample_pipeline.add_stage('Dev Data Generator')
        assert len(sample_pipeline.stages) == 3
        sample_pipeline.stages[-1].instance_name = 'DevDataGenerator_02'

        # Assert that the stage exists in Platform
        sch.publish_pipeline(sample_pipeline)
        sample_pipeline = sch.pipelines.get(pipeline_id=sample_pipeline.pipeline_id)
        assert len(sample_pipeline.stages) == 3
        sample_pipeline.stages[-1].instance_name = 'DevDataGenerator_02'

        # Add another stage to make sure pipelines pulled in from SCH are editable
        sample_pipeline.add_stage('Dev Data Generator')
        assert len(sample_pipeline.stages) == 4
        sample_pipeline.stages[-1].instance_name = 'DevDataGenerator_03'

    except Exception as e:
        raise e
    finally:
        sch.delete_pipeline(sample_pipeline)


def test_add_and_connect_stage_to_linear_pipeline(sch, sch_authoring_sdc_id):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    data_parser = pipeline_builder.add_stage('Data Parser')
    trash = pipeline_builder.add_stage('Trash')
    data_parser >> trash
    sample_pipeline = pipeline_builder.build('sample_pipeline_{}'.format(get_random_string()))
    sch.publish_pipeline(sample_pipeline)

    try:
        # Assert that the inital stages are present
        assert len(sample_pipeline.stages) == 2
        sample_pipeline.stages[0].instance_name = 'DataParser_01'
        assert len(sample_pipeline.stages[0]._data['inputLanes']) == 0
        sample_pipeline.stages[1].instance_name = 'Trash_01'

        # Assert that the stage has been added in memory
        dev_data_generator = sample_pipeline.add_stage('Dev Data Generator')
        dev_data_generator >> sample_pipeline.stages[0]

        assert len(sample_pipeline.stages) == 3
        sample_pipeline.stages[-1].instance_name = 'DevDataGenerator_01'

        data_parser, dev_data_generator = sample_pipeline.stages[0], sample_pipeline.stages[-1]
        assert dev_data_generator._data['outputLanes'][0].split('OutputLane')[0] == dev_data_generator.instance_name
        assert data_parser._data['inputLanes'][0] == dev_data_generator._data['outputLanes'][0]

        # Assert that the stage exists in Platform
        sch.publish_pipeline(sample_pipeline)
        sample_pipeline = sch.pipelines.get(pipeline_id=sample_pipeline.pipeline_id)

        assert len(sample_pipeline.stages) == 3
        sample_pipeline.stages[-1].instance_name = 'DevDataGenerator_01'

        data_parser, dev_data_generator = sample_pipeline.stages[1], sample_pipeline.stages[0]
        print(data_parser)
        print(dev_data_generator)
        assert dev_data_generator._data['outputLanes'][0].split('OutputLane')[0] == dev_data_generator.instance_name
        assert data_parser._data['inputLanes'][0] == dev_data_generator._data['outputLanes'][0]

    except Exception as e:
        raise e
    finally:
        sch.delete_pipeline(sample_pipeline)


def test_add_multiple_destinations_to_linear_pipeline(sch, sch_authoring_sdc_id):
    # Build pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    data_parser = pipeline_builder.add_stage('Data Parser')
    trash = pipeline_builder.add_stage('Trash')

    dev_data_generator >> data_parser
    data_parser >> trash
    sample_pipeline = pipeline_builder.build('sample_pipeline_{}'.format(get_random_string()))
    sch.publish_pipeline(sample_pipeline)

    try:
        # Assert that the initial stages are present
        assert len(sample_pipeline.stages) == 3

        # Assert that the stage has been added in memory
        trash_2 = sample_pipeline.add_stage('Trash')
        assert trash_2.instance_name == 'Trash_02'
        sample_pipeline.stages[-3] >> trash_2

        assert len(sample_pipeline.stages) == 4
        data_parser, trash_1, trash_2 = (
            sample_pipeline.stages[-3],
            sample_pipeline.stages[-2],
            sample_pipeline.stages[-1],
        )

        assert len(data_parser._data['outputLanes']) == 1
        assert len(trash_1._data['inputLanes']) == 1
        assert len(trash_2._data['inputLanes']) == 1
        assert data_parser._data['outputLanes'][0] == trash_1._data['inputLanes'][0]
        assert data_parser._data['outputLanes'][0] == trash_2._data['inputLanes'][0]

        # Assert that the stage exists in Platform
        sch.publish_pipeline(sample_pipeline)
        sample_pipeline = sch.pipelines.get(pipeline_id=sample_pipeline.pipeline_id)

        data_parser, trash_1, trash_2 = (
            sample_pipeline.stages[-3],
            sample_pipeline.stages[-2],
            sample_pipeline.stages[-1],
        )

        assert len(data_parser._data['outputLanes']) == 1
        assert len(trash_1._data['inputLanes']) == 1
        assert len(trash_2._data['inputLanes']) == 1
        assert data_parser._data['outputLanes'][0] == trash_1._data['inputLanes'][0]
        assert data_parser._data['outputLanes'][0] == trash_2._data['inputLanes'][0]

    except Exception as e:
        raise e
    finally:
        sch.delete_pipeline(sample_pipeline)


def test_pipeline_add_stage_with_stage_that_doesnt_exist(sch, sample_pipeline):
    # Assert that the inital stages are present
    assert len(sample_pipeline.stages) == 2
    sample_pipeline.stages[0].instance_name = 'DevDataGenerator_01'
    sample_pipeline.stages[1].instance_name = 'Trash_01'

    # Assert that adding a stage that doesn't exist throws an exception
    with pytest.raises(Exception):
        sample_pipeline.add_stage('Stage that doesnt exist')


def test_add_stage_to_pipeline_in_v1_draft_mode(sch, sch_authoring_sdc_id):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    pipeline = pipeline_builder.build('simple pipeline draft')

    # publish as v1-draft
    sch.publish_pipeline(pipeline, draft=True)
    try:
        # pull draft pipeline and add stages
        pipeline = sch.pipelines.get(pipeline_id=pipeline.pipeline_id, draft=True)
        dev_data_generator = pipeline.add_stage('Dev Data Generator')
        trash = pipeline.add_stage('Trash')
        dev_data_generator >> trash

        # publish pipeline as v1
        sch.publish_pipeline(pipeline)

        # assert that the changes carried over to control hub
        pipeline = sch.pipelines.get(pipeline_id=pipeline.pipeline_id)
        assert len(pipeline.stages) == 2
        assert pipeline.stages[0].instance_name == 'DevDataGenerator_01'
        assert pipeline.stages[1].instance_name == 'Trash_01'

    except Exception as e:
        raise e
    finally:
        sch.delete_pipeline(pipeline)


def test_library_definitions_lazy_loading(sch, sch_authoring_sdc_id):
    # Creating a new pipeline to ensure it hasn't been loaded in-memory yet
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    trash = pipeline_builder.add_stage('Trash')
    dev_data_generator >> trash
    pipeline = pipeline_builder.build('sample_pipeline_{}'.format(get_random_string()))
    sch.publish_pipeline(pipeline)

    assert pipeline.library_definitions is not None
    assert pipeline.library_definitions == json.loads(pipeline._data['libraryDefinitions'])
    assert pipeline._library_definitions == json.loads(pipeline._data['libraryDefinitions'])
    assert pipeline._data_internal['libraryDefinitions'] == pipeline._data['libraryDefinitions']


def test_pipeline_configuration_property(sch, sch_authoring_sdc_id):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    pipeline = pipeline_builder.build('TEST {}'.format(get_random_string()))

    try:
        sch.publish_pipeline(pipeline)
        pipeline = sch.pipelines.get(name=pipeline.name)

        # Picking three arbitrary configuration fields that should exist and checking those values
        assert pipeline.configuration.get('shouldRetry') == pipeline.configuration.retry_pipeline_on_error
        assert (
            pipeline.configuration.get('shouldCreateFailureSnapshot') == pipeline.configuration.create_failure_snapshot
        )
        assert pipeline.configuration.get('triggerInterval') == pipeline.configuration.trigger_interval_in_millis

        # Store old values
        retry_pipeline_on_error = pipeline.configuration.retry_pipeline_on_error
        create_failure_snapshot = pipeline.configuration.create_failure_snapshot
        trigger_interval_in_millis = pipeline.configuration.trigger_interval_in_millis

        # Changing values via the property names
        pipeline.configuration.retry_pipeline_on_error = not pipeline.configuration.retry_pipeline_on_error
        pipeline.configuration.create_failure_snapshot = not pipeline.configuration.create_failure_snapshot
        pipeline.configuration.trigger_interval_in_millis = pipeline.configuration.trigger_interval_in_millis + 1

        # Assert value has been changed in property
        assert pipeline.configuration.retry_pipeline_on_error == (not retry_pipeline_on_error)
        assert pipeline.configuration.create_failure_snapshot == (not create_failure_snapshot)
        assert pipeline.configuration.trigger_interval_in_millis == trigger_interval_in_millis + 1

        # assert that the change has been made in the underlying json
        assert pipeline.configuration.get('shouldRetry') == pipeline.configuration.retry_pipeline_on_error
        assert (
            pipeline.configuration.get('shouldCreateFailureSnapshot') == pipeline.configuration.create_failure_snapshot
        )
        assert pipeline.configuration.get('triggerInterval') == pipeline.configuration.trigger_interval_in_millis

        # Publish pipeline and make sure these changes have been added to Platform
        sch.publish_pipeline(pipeline)
        pipeline = sch.pipelines.get(name=pipeline.name)

        # assert that the change has been made in the underlying json
        assert pipeline.configuration.get('shouldRetry') == (not retry_pipeline_on_error)
        assert pipeline.configuration.get('shouldCreateFailureSnapshot') == (not create_failure_snapshot)
        assert pipeline.configuration.get('triggerInterval') == trigger_interval_in_millis + 1

    except Exception as e:
        raise e

    finally:
        sch.delete_pipeline(pipeline)


def test_stage_configuration_property(sch, sch_authoring_sdc_id):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    pipeline_builder.add_stage('Dev Data Generator')
    pipeline_builder.add_stage('Data Parser')
    pipeline_builder.add_stage('JDBC Multitable Consumer')
    pipeline = pipeline_builder.build('TEST {}'.format(get_random_string()))

    try:
        sch.publish_pipeline(pipeline)
        pipeline = sch.pipelines.get(name=pipeline.name)
        stages = pipeline.stages
        dev_data_gen, data_parser, jdbc_multitable_consumer = stages[0], stages[1], stages[2]

        # Picking three arbitrary configuration fields that should exist and checking those values
        assert dev_data_gen.configuration.get('batchSize') == dev_data_gen.configuration.batch_size
        assert (
            jdbc_multitable_consumer.configuration.get('commonSourceConfigBean.noMoreDataEventDelay')
            == jdbc_multitable_consumer.configuration.no_more_data_event_generation_delay_in_seconds
        )
        assert data_parser.configuration.get('stageOnRecordError') == data_parser.configuration.on_record_error

        # Store older value
        batch_size = dev_data_gen.configuration.batch_size
        no_more_data_event_generation_delay_in_seconds = (
            jdbc_multitable_consumer.configuration.no_more_data_event_generation_delay_in_seconds
        )

        # Changing values via the property names
        dev_data_gen.configuration.batch_size = dev_data_gen.configuration.batch_size + 1
        jdbc_multitable_consumer.configuration.no_more_data_event_generation_delay_in_seconds = (
            jdbc_multitable_consumer.configuration.no_more_data_event_generation_delay_in_seconds + 1
        )
        data_parser.configuration.on_record_error = 'RANDOM VALUE'

        # assert value has been changed
        assert batch_size + 1 == dev_data_gen.configuration.batch_size
        assert no_more_data_event_generation_delay_in_seconds + 1 == (
            jdbc_multitable_consumer.configuration.no_more_data_event_generation_delay_in_seconds
        )
        assert 'RANDOM VALUE' == data_parser.configuration.on_record_error

        # assert that the change has been made in the underlying json
        assert dev_data_gen.configuration.get('batchSize') == dev_data_gen.configuration.batch_size
        assert jdbc_multitable_consumer.configuration.get('commonSourceConfigBean.noMoreDataEventDelay') == (
            jdbc_multitable_consumer.configuration.no_more_data_event_generation_delay_in_seconds
        )
        assert data_parser.configuration.get('stageOnRecordError') == data_parser.configuration.on_record_error

        # Publish pipeline and make sure these changes have been added to Platform
        sch.publish_pipeline(pipeline)
        pipeline = sch.pipelines.get(name=pipeline.name)

        # assert that the change has been made in the underlying json
        assert dev_data_gen.configuration.get('batchSize') == batch_size + 1
        assert jdbc_multitable_consumer.configuration.get('commonSourceConfigBean.noMoreDataEventDelay') == (
            no_more_data_event_generation_delay_in_seconds + 1
        )
        assert data_parser.configuration.get('stageOnRecordError') == 'RANDOM VALUE'

    except Exception as e:
        raise e

    finally:
        sch.delete_pipeline(pipeline)


def test_invalid_pipeline_configuration_property(sch, sch_authoring_sdc_id):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    pipeline = pipeline_builder.build('TEST {}'.format(get_random_string()))

    try:
        sch.publish_pipeline(pipeline)
        pipeline = sch.pipelines.get(name=pipeline.name)

        # Assert that invoking an attribute that does not exist raises an error
        with pytest.raises(AttributeError):
            pipeline.configuration.property_that_clearly_does_not_exist

    except Exception as e:
        raise e

    finally:
        sch.delete_pipeline(pipeline)


def test_invalid_stage_configuration_property(sch, sch_authoring_sdc_id):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    pipeline_builder.add_stage('Dev Data Generator')
    pipeline = pipeline_builder.build('TEST {}'.format(get_random_string()))

    try:
        sch.publish_pipeline(pipeline)
        pipeline = sch.pipelines.get(name=pipeline.name)
        stages = pipeline.stages
        dev_data_gen = stages[0]

        # Assert that invoking an attribute that does not exist raises an error
        with pytest.raises(AttributeError):
            dev_data_gen.configuration.property_that_clearly_does_not_exist

    except Exception as e:
        raise e

    finally:
        sch.delete_pipeline(pipeline)


def test_import_pipeline_and_build_basic_sdc(sch, sch_authoring_sdc_id, sample_pipeline):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    pipeline_builder.import_pipeline(sample_pipeline)
    pipeline = pipeline_builder.build(build_from_imported=True)

    assert pipeline.name == 'Pipeline'
    assert pipeline.version == sample_pipeline.version
    assert pipeline.description == sample_pipeline.description
    assert pipeline.draft == sample_pipeline.draft
    assert pipeline.execution_mode == sample_pipeline.execution_mode

    pipeline_stages, sample_pipeline_stages = pipeline.stages, sample_pipeline.stages
    assert len(pipeline_stages) == len(sample_pipeline_stages)

    for index, stage in enumerate(pipeline_stages):
        assert stage.stage_name == sample_pipeline_stages[index].stage_name


def test_import_pipeline_and_build_basic_transformer(sch, sch_authoring_transformer_id, sample_pipeline):
    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)
    pipeline_builder.import_pipeline(sample_pipeline)
    pipeline = pipeline_builder.build(build_from_imported=True)

    assert pipeline.name == 'Pipeline'
    assert pipeline.version == sample_pipeline.version
    assert pipeline.description == sample_pipeline.description
    assert pipeline.draft == sample_pipeline.draft
    assert pipeline.execution_mode == sample_pipeline.execution_mode

    pipeline_stages, sample_pipeline_stages = pipeline.stages, sample_pipeline.stages
    assert len(pipeline_stages) == len(sample_pipeline_stages)

    for index, stage in enumerate(pipeline_stages):
        assert stage.stage_name == sample_pipeline_stages[index].stage_name


def test_import_pipeline_with_new_title(sch, sch_authoring_sdc_id, sample_pipeline):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    pipeline_builder.import_pipeline(sample_pipeline)
    pipeline = pipeline_builder.build('TEST IMPORT NAME', build_from_imported=True)

    assert pipeline.name == 'TEST IMPORT NAME'


def test_import_pipeline_with_new_title_transformer(sch, sch_authoring_transformer_id, sample_pipeline):
    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)
    pipeline_builder.import_pipeline(sample_pipeline)
    pipeline = pipeline_builder.build('TEST IMPORT NAME', build_from_imported=True)

    assert pipeline.name == 'TEST IMPORT NAME'


def test_import_pipeline_and_build_with_commit_id_regeneration_false_sdc(sch, sch_authoring_sdc_id, sample_pipeline):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    pipeline_builder.import_pipeline(sample_pipeline, commit_id_regeneration=False)
    pipeline = pipeline_builder.build(build_from_imported=True)

    assert pipeline.name == 'Pipeline'
    assert pipeline.version == sample_pipeline.version
    assert pipeline.description == sample_pipeline.description
    assert pipeline.draft == sample_pipeline.draft
    assert pipeline.execution_mode == sample_pipeline.execution_mode

    pipeline_stages, sample_pipeline_stages = pipeline.stages, sample_pipeline.stages
    assert len(pipeline_stages) == len(sample_pipeline_stages)

    for index, stage in enumerate(pipeline_stages):
        assert stage.stage_name == sample_pipeline_stages[index].stage_name

    assert pipeline.commit_id == sample_pipeline.commit_id
    assert pipeline.commit_message == sample_pipeline.commit_message
    assert pipeline.commit_time == sample_pipeline.commit_time


def test_import_pipeline_and_build_with_commit_id_regeneration_false_transformer(
    sch, sch_authoring_transformer_id, sample_pipeline
):
    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)
    pipeline_builder.import_pipeline(sample_pipeline, commit_id_regeneration=False)
    pipeline = pipeline_builder.build(build_from_imported=True)

    assert pipeline.name == 'Pipeline'
    assert pipeline.version == sample_pipeline.version
    assert pipeline.description == sample_pipeline.description
    assert pipeline.draft == sample_pipeline.draft
    assert pipeline.execution_mode == sample_pipeline.execution_mode

    pipeline_stages, sample_pipeline_stages = pipeline.stages, sample_pipeline.stages
    assert len(pipeline_stages) == len(sample_pipeline_stages)

    for index, stage in enumerate(pipeline_stages):
        assert stage.stage_name == sample_pipeline_stages[index].stage_name

    assert pipeline.commit_id == sample_pipeline.commit_id
    assert pipeline.commit_message == sample_pipeline.commit_message
    assert pipeline.commit_time == sample_pipeline.commit_time


def test_import_pipeline_and_build_with_regenerate_id_false_sdc(sch, sch_authoring_sdc_id, sample_pipeline):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    pipeline_builder.import_pipeline(sample_pipeline, regenerate_id=False)
    pipeline = pipeline_builder.build(build_from_imported=True)

    assert pipeline.name == 'Pipeline'
    assert pipeline.version == sample_pipeline.version
    assert pipeline.description == sample_pipeline.description
    assert pipeline.draft == sample_pipeline.draft
    assert pipeline.execution_mode == sample_pipeline.execution_mode

    pipeline_stages, sample_pipeline_stages = pipeline.stages, sample_pipeline.stages
    assert len(pipeline_stages) == len(sample_pipeline_stages)

    for index, stage in enumerate(pipeline_stages):
        assert stage.stage_name == sample_pipeline_stages[index].stage_name

    assert pipeline.pipeline_id == sample_pipeline.pipeline_id


def test_import_pipeline_and_build_with_regenerate_id_false_transformer(
    sch, sch_authoring_transformer_id, sample_pipeline
):
    pipeline_builder = sch.get_pipeline_builder(engine_type='transformer', engine_id=sch_authoring_transformer_id)
    pipeline_builder.import_pipeline(sample_pipeline, regenerate_id=False)
    pipeline = pipeline_builder.build(build_from_imported=True)

    assert pipeline.name == 'Pipeline'
    assert pipeline.version == sample_pipeline.version
    assert pipeline.description == sample_pipeline.description
    assert pipeline.draft == sample_pipeline.draft
    assert pipeline.execution_mode == sample_pipeline.execution_mode

    pipeline_stages, sample_pipeline_stages = pipeline.stages, sample_pipeline.stages
    assert len(pipeline_stages) == len(sample_pipeline_stages)

    for index, stage in enumerate(pipeline_stages):
        assert stage.stage_name == sample_pipeline_stages[index].stage_name

    assert pipeline.pipeline_id == sample_pipeline.pipeline_id


def test_import_pipeline_with_invalid_pipeline_type(sch, sch_authoring_sdc_id):
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    with pytest.raises(ValueError):
        pipeline_builder.import_pipeline('sample_pipeline')


def test_pipeline_uses_same_builder(sample_pipeline):
    sample_pipeline.add_stage("Trash")
    sample_pipeline_builder_1 = sample_pipeline._builder
    sample_pipeline.add_stage("Trash")
    assert sample_pipeline_builder_1 is sample_pipeline._builder


def test_pipeline_uses_same_builder_with_engine_changes(sch, sdc_deployment):
    pipeline = None
    try:
        deployment = sdc_deployment
        pipeline_name = 'sample_pipeline_{}'.format(get_random_string())
        engine = deployment.registered_engines[0]
        engine_id = engine.id
        engine_type = "data_collector"
        stage_to_add = "Cassandra"
        library_to_add = "cassandra_3"

        # Create a pipeline
        pipeline_builder = sch.get_pipeline_builder(engine_type=engine_type, engine_id=engine_id)
        pipeline = pipeline_builder.build(pipeline_name)
        sch.publish_pipeline(pipeline)

        # Retrieve a pipeline and check it has no builder
        pipeline = sch.pipelines.get(name=pipeline_name)
        assert pipeline._builder is None

        # Check a pipelineBuilder is made
        pipeline.add_stage("Trash")
        assert pipeline._builder is not None
        pipeline_builder_1 = pipeline._builder

        # Check pipeline can't add stage that does not exist
        with pytest.raises(Exception) as e:
            pipeline.add_stage(stage_to_add)
        assert str(e.value) == "Could not find stage ({}).".format(stage_to_add)

        # Update the deployment and engine with a new library
        deployment.engine_configuration.stage_libs.extend([library_to_add])
        sch.update_deployment(deployment)
        sch.restart_engines(engine)
        # Wait for engine to restart
        time.sleep(300)

        # Check ability to add new stage with the same builder object
        pipeline_stage_len = len(pipeline.stages)
        pipeline.add_stage(stage_to_add)
        assert len(pipeline.stages) == pipeline_stage_len + 1
        assert pipeline._builder is pipeline_builder_1
    finally:
        if pipeline:
            sch.delete_pipeline(pipeline)


@pytest.mark.parametrize(
    "sch_authoring_id, engine_type",
    [("sch_authoring_sdc_id", "transformer"), ("sch_authoring_transformer_id", "data_collector")],
)
def test_call_get_pipeline_builder_with_disparate_engine_type_and_engine(sch, sch_authoring_id, engine_type, request):
    sch_authoring_id = request.getfixturevalue(sch_authoring_id)
    with pytest.raises(TypeError):
        sch.get_pipeline_builder(engine_type, sch_authoring_id)



# Copyright 2022 StreamSets Inc.

# fmt: off
import json
import os
from datetime import datetime

import pytest

from streamsets.sdk.sch_models import ACL, Configuration, Pipeline
from streamsets.sdk.st_models import Preview as StPreview
from streamsets.sdk.utils import SeekableList, get_random_string

# fmt: on

NUM_PIPELINES = 12


@pytest.fixture(scope="module")
def sample_pipelines(sch, snowflake_config):
    """A set of trivial pipelines:

    dev_data_generator >> trash
    """
    pipelines = []

    for i in range(NUM_PIPELINES):
        pipeline_builder = sch.get_pipeline_builder(engine_type='snowflake')

        snowflake_table_name = 'com_streamsets_transformer_snowpark_origin_snowflake_table_TableDOrigin'
        snowflake_table = pipeline_builder.add_stage(name=snowflake_table_name).set_attributes(table='dinos')
        trash = pipeline_builder.add_stage('Trash')
        snowflake_table >> trash

        pipeline_name = 'pipeline_snowflake_test_{}_{}'.format(i, get_random_string())
        pipeline = pipeline_builder.build(pipeline_name)
        pipeline.configuration['connectionString'] = snowflake_config['connectionString']
        pipeline.configuration['db'] = snowflake_config['db']
        pipeline.configuration['warehouse'] = snowflake_config['warehouse']
        pipeline.configuration['schema'] = snowflake_config['schema']
        sch.publish_pipeline(pipeline)

        pipelines.append(pipeline)

    try:
        yield pipelines
    finally:
        for pipeline in pipelines:
            sch.delete_pipeline(pipeline)


@pytest.fixture(scope="module")
def sample_pipeline(sch, snowflake_config):
    """A trivial pipelines:

    dev_data_generator >> trash
    """
    pipeline_builder = sch.get_pipeline_builder(engine_type='snowflake')

    snowflake_table_name = 'com_streamsets_transformer_snowpark_origin_snowflake_table_TableDOrigin'
    snowflake_table = pipeline_builder.add_stage(name=snowflake_table_name).set_attributes(table='dinos')
    trash = pipeline_builder.add_stage('Trash')
    snowflake_table >> trash
    pipeline = pipeline_builder.build('sample_pipeline_{}'.format(get_random_string()))
    pipeline.configuration['connectionString'] = snowflake_config['connectionString']
    pipeline.configuration['db'] = snowflake_config['db']
    pipeline.configuration['warehouse'] = snowflake_config['warehouse']
    pipeline.configuration['schema'] = snowflake_config['schema']
    sch.publish_pipeline(pipeline)

    try:
        yield pipeline
    finally:
        sch.delete_pipeline(pipeline)


@pytest.mark.snowflake
def test_iter(sch, sample_pipelines):
    iter_results = []
    for pipeline in sch.pipelines:
        iter_results.append(pipeline.pipeline_id)
    assert len(iter_results) >= len(sample_pipelines)
    assert not ({pipeline.pipeline_id for pipeline in sample_pipelines} - set(iter_results))


@pytest.mark.snowflake
def test_pipelines_get_returns_pipeline_object(sch, sample_pipelines):
    assert sample_pipelines[0] in sch.pipelines
    assert isinstance(sch.pipelines.get(organization=sch.organization), Pipeline)


@pytest.mark.snowflake
def test_pipeline_in_returns_true(sch, sample_pipelines):
    assert sample_pipelines[0] in sch.pipelines


@pytest.mark.snowflake
def test_pipeline_contains(sch, sample_pipelines):
    pipeline = sample_pipelines[0]
    assert sch.pipelines.contains(pipeline_id=pipeline.pipeline_id)
    assert sch.pipelines.contains(name=pipeline.name)
    assert not sch.pipelines.contains(pipeline_id='impossible_to_clash_with_this_zbsdudcugeqwgui')


@pytest.mark.snowflake
def test_pipelines_len_works(sch, sample_pipelines):
    assert len(sch.pipelines) >= len(sample_pipelines)


# Tests for __getitem__
@pytest.mark.snowflake
def test_pipelines_getitem_int_pos_invalid_index(sch):
    total_number_of_pipelines = len(sch.pipelines)
    with pytest.raises(IndexError):
        sch.pipelines[total_number_of_pipelines]


@pytest.mark.snowflake
def test_pipelines_getitem_int_neg_invalid_index(sch):
    total_number_of_pipelines = len(sch.pipelines)
    with pytest.raises(IndexError):
        sch.pipelines[-total_number_of_pipelines - 1]


@pytest.mark.snowflake
def test_pipelines_getitem_int_instance_type(sch, sample_pipelines):
    total_number_of_pipelines = len(sch.pipelines)
    assert isinstance(sch.pipelines[total_number_of_pipelines - 1], Pipeline)


@pytest.mark.snowflake
@pytest.mark.skip('This test will fail until TLKT-1052 is resolved')
def test_pipelines_getitem_slice_all(sch):
    total_number_of_pipelines = len(sch.pipelines)
    assert len(sch.pipelines[:]) == total_number_of_pipelines


@pytest.mark.snowflake
@pytest.mark.skip('This test will fail until TLKT-1052 is resolved')
def test_pipelines_getitem_slice_stop(sch):
    assert len(sch.pipelines[:NUM_PIPELINES]) == NUM_PIPELINES


@pytest.mark.snowflake
@pytest.mark.skip('This test will fail until TLKT-1052 is resolved')
def test_pipelines_getitem_slice_start_stop(sch):
    assert len(sch.pipelines[0:NUM_PIPELINES]) == NUM_PIPELINES


@pytest.mark.snowflake
@pytest.mark.skip('This test will fail until TLKT-1052 is resolved')
def test_pipelines_getitem_slice_neg_start(sch):
    total_number_of_pipelines = len(sch.pipelines)
    assert len(sch.pipelines[-total_number_of_pipelines:]) == total_number_of_pipelines


@pytest.mark.snowflake
@pytest.mark.skip('This test will fail until TLKT-1052 is resolved')
def test_pipelines_getitem_slice_neg_start_0_stop(sch):
    total_number_of_pipelines = len(sch.pipelines)
    # stop is 0 so, it should always return empty list
    assert len(sch.pipelines[-total_number_of_pipelines:0]) == 0


@pytest.mark.snowflake
@pytest.mark.skip('This test will fail until TLKT-1052 is resolved')
def test_pipelines_getitem_slice_extra_neg_start(sch):
    total_number_of_pipelines = len(sch.pipelines)
    # eg. [1,2,3,4][-9:] = [1,2,3,4]
    assert len(sch.pipelines[-total_number_of_pipelines - 5 :]) == total_number_of_pipelines


@pytest.mark.snowflake
@pytest.mark.skip('This test will fail until TLKT-1052 is resolved')
def test_pipelines_getitem_slice_step_size(sch):
    total_number_of_pipelines = len(sch.pipelines)
    # verify with step size for both the cases of total_number_of_pipelines being an odd or even number
    assert len(sch.pipelines[::2]) == total_number_of_pipelines // 2 + total_number_of_pipelines % 2


@pytest.mark.snowflake
@pytest.mark.skip('This test will fail until TLKT-1052 is resolved')
def test_pipelines_getitem_slice_neg_step_size(sch):
    # verify negative step size
    assert len(sch.pipelines[-1:-2:-1]) == 1


# Tests for get_all
@pytest.mark.snowflake
def test_get_all_returns_seekable_list(sch):
    assert isinstance(sch.pipelines.get_all(), SeekableList)


@pytest.mark.snowflake
def test_get_all_filter_by_organization_and_version(sch):
    assert sch.pipelines.get_all(organization=sch.organization, version='1')


@pytest.mark.snowflake
def test_get_all_filter_by_text(sch, sample_pipelines):
    pipelines = sch.pipelines.get_all(filter_text='pipeline_snowflake_test_')
    assert len(pipelines) == NUM_PIPELINES


# ACL tests
@pytest.mark.snowflake
def test_pipeline_acl_get(sch):
    pipeline = sch.pipelines[0]
    acl = pipeline.acl
    assert isinstance(acl, ACL)
    assert acl.resource_type == 'PIPELINE'


@pytest.mark.snowflake
def test_import_export_pipelines(sch, snowflake_config):
    # Create a pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='snowflake')

    snowflake_table_name = 'com_streamsets_transformer_snowpark_origin_snowflake_table_TableDOrigin'
    snowflake_table = pipeline_builder.add_stage(name=snowflake_table_name).set_attributes(table='dinos')
    trash = pipeline_builder.add_stage('Trash')
    snowflake_table >> trash
    pipeline = pipeline_builder.build('import_export_pipeline')
    pipeline.configuration['connectionString'] = snowflake_config['connectionString']
    pipeline.configuration['db'] = snowflake_config['db']
    pipeline.configuration['warehouse'] = snowflake_config['warehouse']
    pipeline.configuration['schema'] = snowflake_config['schema']
    sch.publish_pipeline(pipeline)

    pipelines_zip_data = sch.export_pipelines([pipeline])
    assert isinstance(pipelines_zip_data, bytes)

    path = '/tmp'
    filename = 'import_export_pipelines_{}.zip'.format(datetime.now().strftime('%Y_%m_%d_%H_%M_%S'))
    with open('{}/{}'.format(path, filename), 'wb') as pipelines_zip_file:
        pipelines_zip_file.write(pipelines_zip_data)
    pipelines_zip_file_name = '{}/{}'.format(path, filename)

    try:
        pipelines_before = [pipeline]
        sch.delete_pipeline(pipeline)
        with open(pipelines_zip_file_name, 'rb') as input_file:
            pipelines_after = sch.import_pipelines_from_archive(input_file, 'Test Import Pipelines')
        assert len(pipelines_before) == len(pipelines_after)
    except Exception as ex:
        raise ex
    finally:
        os.remove(pipelines_zip_file_name)
        if pipeline:
            sch.delete_pipeline(pipeline)


@pytest.mark.snowflake
def test_get_pipeline_templates(sch):
    templates = sch.pipelines.get_all(template=True)
    assert isinstance(templates, SeekableList)
    if templates:
        assert isinstance(templates[0], Pipeline)


@pytest.mark.snowflake
def test_update_pipeline(sch, snowflake_config):
    # Build and publish pipeline
    pipeline_builder = sch.get_pipeline_builder(engine_type='snowflake')

    snowflake_table_name = 'com_streamsets_transformer_snowpark_origin_snowflake_table_TableDOrigin'
    snowflake_table = pipeline_builder.add_stage(name=snowflake_table_name).set_attributes(table='dinos')
    trash = pipeline_builder.add_stage('Trash')
    snowflake_table >> trash
    pipeline = pipeline_builder.build('pipeline_to_be_updated')
    pipeline.configuration['connectionString'] = snowflake_config['connectionString']
    pipeline.configuration['db'] = snowflake_config['db']
    pipeline.configuration['warehouse'] = snowflake_config['warehouse']
    pipeline.configuration['schema'] = snowflake_config['schema']
    sch.publish_pipeline(pipeline)

    try:
        # Update batch size
        for stage in pipeline.stages:
            if stage.stage_name == 'com_streamsets_transformer_snowpark_origin_snowflake_table_TableDOrigin':
                assert stage.label != 'Updated stage name'
                stage.label = 'Updated stage name'
        sch.publish_pipeline(pipeline)

        # Confirm pipeline.stages is updated
        for stage in pipeline.stages:
            if stage.stage_name == 'com_streamsets_transformer_snowpark_origin_snowflake_table_TableDOrigin':
                assert stage.label == 'Updated stage name'

        # Confirm it is updated in json data
        for stage in json.loads(pipeline._data['pipelineDefinition'])['stages']:
            if stage['stageName'] == 'com_streamsets_transformer_snowpark_origin_snowflake_table_TableDOrigin':
                for config in stage['configuration']:
                    if config['name'] == 'label':
                        assert config['value'] == 'Updated stage name'
    finally:
        sch.delete_pipeline(pipeline)


@pytest.mark.snowflake
def test_execution_mode_config(sch, snowflake_config):
    pipeline_builder = sch.get_pipeline_builder(engine_type='snowflake')

    snowflake_table_name = 'com_streamsets_transformer_snowpark_origin_snowflake_table_TableDOrigin'
    snowflake_table = pipeline_builder.add_stage(name=snowflake_table_name).set_attributes(table='dinos')
    trash = pipeline_builder.add_stage('Trash')
    snowflake_table >> trash
    pipeline = pipeline_builder.build('execution_mode_config')
    pipeline.configuration['connectionString'] = snowflake_config['connectionString']
    pipeline.configuration['db'] = snowflake_config['db']
    pipeline.configuration['warehouse'] = snowflake_config['warehouse']
    pipeline.configuration['schema'] = snowflake_config['schema']
    sch.publish_pipeline(pipeline)
    try:
        pipeline_definition = json.loads(pipeline._data['pipelineDefinition'])
        assert 'executionMode' not in pipeline_definition
        configuration = Configuration(configuration=pipeline_definition['configuration'])
        assert configuration.get('executionMode', None) is not None
    finally:
        sch.delete_pipeline(pipeline)


@pytest.mark.snowflake
def test_duplicate_pipeline(sch, sample_pipeline):
    duplicated_pipelines = sch.duplicate_pipeline(sample_pipeline)
    try:
        assert isinstance(duplicated_pipelines, SeekableList)
        assert len(duplicated_pipelines) == 1
        assert (
            sch.pipelines.get(name='{} copy'.format(sample_pipeline.name), only_published=False).pipeline_id
            == duplicated_pipelines[0].pipeline_id
        )
    finally:
        for pipeline in duplicated_pipelines:
            sch.delete_pipeline(pipeline)


@pytest.mark.snowflake
def test_duplicate_pipelines(sch, sample_pipeline):
    name = 'test_duplicate_pipeline_snowflake_{}_'.format(get_random_string())
    duplicated_pipelines = sch.duplicate_pipeline(
        sample_pipeline, name=name, description='Testing Duplication', number_of_copies=3
    )

    pipeline_id_to_object = {}
    for pipeline in duplicated_pipelines:
        pipeline_id_to_object[pipeline.pipeline_id] = pipeline

    try:
        assert isinstance(duplicated_pipelines, SeekableList)
        assert len(duplicated_pipelines) == 3

        for duplicate_pipeline in duplicated_pipelines:
            pipeline = sch.pipelines.get(name=duplicate_pipeline.name, only_published=False)

            assert pipeline.pipeline_id in pipeline_id_to_object
            assert pipeline.name == pipeline_id_to_object[pipeline.pipeline_id].name

    finally:
        for pipeline in duplicated_pipelines:
            sch.delete_pipeline(pipeline)


@pytest.mark.snowflake
def test_create_and_publish_draft(sch, snowflake_config):
    pipeline_builder = sch.get_pipeline_builder(engine_type='snowflake')

    snowflake_table_name = 'com_streamsets_transformer_snowpark_origin_snowflake_table_TableDOrigin'
    snowflake_table = pipeline_builder.add_stage(name=snowflake_table_name).set_attributes(table='dinos')
    trash = pipeline_builder.add_stage('Trash')
    snowflake_table >> trash
    pipeline = pipeline_builder.build('simple pipeline draft')
    pipeline.configuration['connectionString'] = snowflake_config['connectionString']
    pipeline.configuration['db'] = snowflake_config['db']
    pipeline.configuration['warehouse'] = snowflake_config['warehouse']
    pipeline.configuration['schema'] = snowflake_config['schema']
    sch.publish_pipeline(pipeline, draft=True)
    try:
        assert pipeline.draft is True
        assert pipeline.name == 'simple pipeline draft'
        assert pipeline.version == '1-DRAFT'
        sch.publish_pipeline(pipeline)
        assert getattr(pipeline, 'draft', False) is False
        assert pipeline.name == 'simple pipeline draft'
    finally:
        sch.delete_pipeline(pipeline)


@pytest.mark.snowflake
def test_create_and_publish_validate(sch, snowflake_config):
    pipeline_builder = sch.get_pipeline_builder(engine_type='snowflake')

    snowflake_table_name = 'com_streamsets_transformer_snowpark_origin_snowflake_table_TableDOrigin'
    pipeline_builder.add_stage(name=snowflake_table_name).set_attributes(table='dinos')
    pipeline = pipeline_builder.build('simple pipeline validate true')
    pipeline.configuration['connectionString'] = snowflake_config['connectionString']
    pipeline.configuration['db'] = snowflake_config['db']
    pipeline.configuration['warehouse'] = snowflake_config['warehouse']
    pipeline.configuration['schema'] = snowflake_config['schema']
    sch.publish_pipeline(pipeline, validate=True)
    try:
        # The published pipeline is not valid since it doesn't have a destination. Checking that it was marked as such
        # when validate is True
        pipeline_definition = json.loads(pipeline.pipeline_definition)
        assert not pipeline_definition['valid']
        # Publishing the pipeline again with validate equals False, to see if now it always shows as valid
        sch.publish_pipeline(pipeline)
        pipeline_definition = json.loads(pipeline.pipeline_definition)
        assert pipeline_definition['valid']
    finally:
        sch.delete_pipeline(pipeline)


@pytest.mark.snowflake
def test_publishing_duplicate_of_pipeline_draft(sch, snowflake_config):
    pipeline_builder = sch.get_pipeline_builder(engine_type='snowflake')

    snowflake_table_name = 'com_streamsets_transformer_snowpark_origin_snowflake_table_TableDOrigin'
    snowflake_table = pipeline_builder.add_stage(name=snowflake_table_name).set_attributes(table='dinos')
    trash = pipeline_builder.add_stage('Trash')
    snowflake_table >> trash
    pipeline = pipeline_builder.build('simple pipeline draft')
    pipeline.configuration['connectionString'] = snowflake_config['connectionString']
    pipeline.configuration['db'] = snowflake_config['db']
    pipeline.configuration['warehouse'] = snowflake_config['warehouse']
    pipeline.configuration['schema'] = snowflake_config['schema']
    sch.publish_pipeline(pipeline, draft=True)
    try:
        duplicated_pipeline = sch.duplicate_pipeline(pipeline)[0]
        assert duplicated_pipeline.draft is True
        assert duplicated_pipeline.version == '1-DRAFT'
        duplicated_pipeline_name = duplicated_pipeline.name
        sch.publish_pipeline(duplicated_pipeline)
        assert getattr(duplicated_pipeline, 'draft', False) is False
        assert duplicated_pipeline.name == duplicated_pipeline_name
    finally:
        sch.delete_pipeline(duplicated_pipeline)
        sch.delete_pipeline(pipeline)


@pytest.fixture(scope="module")
def preview_pipeline(sch, snowflake_config):
    """A spcific pipeline to test preview end point"""
    pipeline_builder = sch.get_pipeline_builder(engine_type='snowflake')

    query = "select COLUMN1 as col1, COLUMN2 as col2 from values (1, 'one'), (2, 'two'), (3, 'three')"
    query_origin = pipeline_builder.add_stage('Snowflake Query', type='origin')
    query_origin.query = query

    filter_processor = pipeline_builder.add_stage('Filter')
    filter_processor.filter_by = "BY_SQL"
    filter_processor.where_clause = "col1 >= 3"

    trash = pipeline_builder.add_stage('Trash')
    query_origin >> filter_processor >> trash

    pipeline = pipeline_builder.build('sample_pipeline_{}'.format(get_random_string()))
    pipeline.configuration['connectionString'] = snowflake_config['connectionString']
    pipeline.configuration['db'] = snowflake_config['db']
    pipeline.configuration['warehouse'] = snowflake_config['warehouse']
    pipeline.configuration['schema'] = snowflake_config['schema']
    sch.publish_pipeline(pipeline)

    try:
        yield pipeline
    finally:
        sch.delete_pipeline(pipeline)


@pytest.mark.snowflake
@pytest.mark.parametrize('push_limit_down', [True, False])
def test_get_pipeline_preview_end_limit(sch, preview_pipeline, push_limit_down):
    sch.publish_pipeline(preview_pipeline)
    preview = sch.run_pipeline_preview(preview_pipeline, batch_size=1, push_limit_down=push_limit_down).preview
    assert preview is not None
    assert preview.issues.issues_count == 0
    assert isinstance(preview, StPreview)



import json

# Referenced Data for function dummy_connection(mocker)
dummy_connection_connection_definition_json = json.dumps(
    {
        'type': 'STREAMSETS_SNOWFLAKE',
        'version': '7',
        'configuration': [
            {'name': 'connectionProperties', 'value': [{}]},
            {'name': 'privateKeyPemPath', 'value': None},
            {'name': 'useSnowflakeRole', 'value': False},
            {'name': 'user', 'value': 'Test User'},
            {'name': 'account', 'value': 'Test Account'},
        ],
    }
)
dummy_connection_library_definitions_json = json.dumps(
    {
        'configDefinitions': [
            {
                'name': 'connectionProperties',
                'label': 'Connection Properties',
                'fieldName': 'connectionProperties',
            },
            {
                'name': 'privateKeyPemPath',
                'label': 'Private Key Path',
                'fieldName': 'privateKeyPemPath',
            },
            {
                'name': 'useSnowflakeRole',
                'label': 'Use Snowflake Role',
                'fieldName': 'useSnowflakeRole',
            },
            {
                'name': 'account',
                'label': 'Account',
                'fieldName': 'account',
            },
            {
                'name': 'User',
                'label': 'user',
                'fieldName': 'user',
            },
        ]
    }
)

DUMMY_CONNECTION_JSON = {
    'id': '011c7781-48a1-4d4c-93c8-1ad42a5505cb:1a1b7cf4-862e-11ed-94d5-41a005fd7967',
    'name': 'Snowflake Demo Connection',
    'connectionType': 'STREAMSETS_SNOWFLAKE',
    'connectionDefinition': dummy_connection_connection_definition_json,
    'libraryDefinition': dummy_connection_library_definitions_json,
}

DUMMY_DESTINATION_STAGE_JSON = {
    'instanceName': 'AmazonS3_1',
    'library': 'streamsets-datacollector-aws-lib',
    'stageName': 'com_streamsets_pipeline_stage_destination_s3_AmazonS3DTarget',
    'stageVersion': '15',
    'configuration': [
        {'name': 's3TargetConfigBean.s3Config.connection.awsConfig.awsSecretAccessKey', 'value': '1212'},
        {'name': 's3TargetConfigBean.tmConfig.threadPoolSize', 'value': 10},
        {'name': 's3TargetConfigBean.tmConfig.multipartUploadThreshold', 'value': 26843545},
        {'name': 's3TargetConfigBean.s3Config.connection.proxyConfig.proxyWorkstation', 'value': None},
        {'name': 's3TargetConfigBean.s3Config.connection.proxyConfig.socketTimeout', 'value': 50},
        {'name': 's3TargetConfigBean.s3Config.connection.endpoint', 'value': None},
        {'name': 's3TargetConfigBean.s3Config.connectionSelection', 'value': 'MANUAL'},
        {'name': 's3TargetConfigBean.s3Config.connection.awsConfig.sessionDuration', 'value': 3600},
        {'name': 'stageOnRecordError', 'value': 'TO_ERROR'},
        {'name': 'stageRequiredFields', 'value': []},
        {'name': 'stageRecordPreconditions', 'value': []},
    ],
    'uiInfo': {
        'colorIcon': 'Destination_Amazon_S3.png',
        'outputStreamLabels': None,
        'yPos': 50,
        'stageType': 'TARGET',
        'rawSource': None,
        'icon': 's3.png',
        'description': '',
        'inputStreamLabels': None,
        'label': 'Amazon S3 1',
        'xPos': 280,
        'displayMode': 'BASIC',
    },
    'inputLanes': [],
    'outputLanes': [],
    'eventLanes': ['AmazonS3_1_EventLane'],
    'services': [
        {
            'service': 'com.streamsets.pipeline.api.service.dataformats.DataFormatGeneratorService',
            'serviceVersion': 3,
            'configuration': [
                {'name': 'dataFormat', 'value': 'BINARY'},
                {'name': 'dataGeneratorFormatConfig.charset', 'value': 'UTF-8'},
                {'name': 'dataGeneratorFormatConfig.csvHeader', 'value': 'NO'},
                {'name': 'dataGeneratorFormatConfig.csvFileFormat', 'value': 'CSV'},
                {'name': 'dataGeneratorFormatConfig.csvCustomDelimiter', 'value': '|'},
                {'name': 'dataGeneratorFormatConfig.avroSchema', 'value': None},
                {'name': 'dataGeneratorFormatConfig.xmlPrettyPrint', 'value': True},
                {'name': 'dataGeneratorFormatConfig.xmlValidateSchema', 'value': False},
                {'name': 'dataGeneratorFormatConfig.xmlSchema', 'value': None},
            ],
        }
    ],
}



# Copyright Streamsets 2023

# fmt: off
import pytest

from streamsets.sdk.sdc_models import PipelineBuilder as SdcPipelineBuilder
from streamsets.sdk.st_models import PipelineBuilder as StPipelineBuilder

from .resources.pipeline_builder_data import CONFIG_TYPE_TEST_CASES_JSON as CONFIG_TYPE_TEST_CASES

# fmt: on


@pytest.mark.parametrize('pipeline_builder', ['sch_sdc_pipeline_builder', 'sch_st_pipeline_builder'])
def test_pipeline_builder_add_stage_stage_exist(pipeline_builder, request):
    pipeline_builder = request.getfixturevalue(pipeline_builder)
    stage_to_add = "Trash"
    pipeline_builder.add_stage(stage_to_add)
    dummy_pipeline_stage = pipeline_builder._pipeline[pipeline_builder._config_key]['stages'][0]
    assert stage_to_add in dummy_pipeline_stage["instanceName"]


@pytest.mark.parametrize('pipeline_builder', ['sch_sdc_pipeline_builder', 'sch_st_pipeline_builder'])
def test_pipeline_builder_add_stage_stage_does_not_exist(pipeline_builder, request):
    pipeline_builder = request.getfixturevalue(pipeline_builder)
    fake_stage_name = "Cannot Exist"
    with pytest.raises(Exception) as e:
        pipeline_builder.add_stage(fake_stage_name)
    assert str(e.value) == "Could not find stage ({}).".format(fake_stage_name)


@pytest.mark.parametrize('pipeline_builder', ['sch_sdc_pipeline_builder', 'sch_st_pipeline_builder'])
def test_add_stage_supported_types(pipeline_builder, request):
    pipeline_builder = request.getfixturevalue(pipeline_builder)
    snowflake_destination = pipeline_builder.add_stage('Snowflake', type='destination')

    assert 'STREAMSETS_SNOWFLAKE' in snowflake_destination.supported_connection_types


@pytest.mark.parametrize('builder_type', ['sdc', 'st'])
@pytest.mark.parametrize('test_case', CONFIG_TYPE_TEST_CASES)
def test_configuration_default_values(builder_type, pipeline_builder_definitions, pipeline_builder_json, test_case):
    config_definition, expected_value = test_case

    if builder_type == 'sdc':
        builder_class = SdcPipelineBuilder
    elif builder_type == 'st':
        builder_class = StPipelineBuilder

    # Assert that the default value for each model type returns the expected value when there is no default set in
    # the root level of the configuration definition.
    pipeline_builder_definitions['stages'][0]['configDefinitions'] = [config_definition.copy()]
    builder = builder_class(pipeline=pipeline_builder_json, definitions=pipeline_builder_definitions)
    stage_instance = builder.add_stage(pipeline_builder_definitions['stages'][0]['label'])
    assert stage_instance.configuration[config_definition['name']] == expected_value

    # Assert that the default value in the root level of the configuration definition will take precedence over
    # all others.
    pipeline_builder_definitions['stages'][0]['configDefinitions'][0]['defaultValue'] = "OVERRIDE"
    builder = builder_class(pipeline=pipeline_builder_json, definitions=pipeline_builder_definitions)
    stage_instance = builder.add_stage(pipeline_builder_definitions['stages'][0]['label'])
    assert stage_instance.configuration[config_definition['name']] == "OVERRIDE"



# fmt: off
import json
import warnings

import pytest
from tests.mocks.mock_api import MockResponse

from streamsets.sdk.constants import SNOWFLAKE_EXECUTOR_TYPE, STATUS_ERRORS
from streamsets.sdk.exceptions import InvalidError
from streamsets.sdk.sch_api import Command
from streamsets.sdk.sch_models import Pipeline
from streamsets.sdk.utils import get_random_string

# fmt: on


class DummyPipelineBuilder:
    """
    A Dummy Pipeline Class Builder
    """

    def __init__(self):
        self._config_key = 'pipelineBuilder'
        self._pipeline = {self._config_key: {'stages': []}}


class MockControlHub:
    def __init__(self):
        self.api_client = MockApiClient()


class MockApiClient:
    def get_pipelines_definitions(self, val):
        return Command(self, MockResponse({"foo": "bar"}, 200))


@pytest.fixture(scope="function")
def dummy_pipeline():
    pipeline_json = {
        'pipelineId': 1,
        'commitId': None,
        'name': 'Test Pipeline',
        'version': 1,
        'sdcId': 'a738b839-bac3-4118-b0f4-eb6509f0b7cf',
    }
    pipeline_definition_json = json.dumps({'title': 'dummy_value'})
    rules_definition_json = {}
    library_definitions_json = json.dumps({'schemaVersion': 1})
    builder = DummyPipelineBuilder()

    return Pipeline(
        pipeline=pipeline_json,
        builder=builder,
        pipeline_definition=pipeline_definition_json,
        rules_definition=rules_definition_json,
        library_definitions=library_definitions_json,
        control_hub=MockControlHub(),
    )


def test_library_definitions_in_data_str_sanity(dummy_pipeline):
    assert isinstance(dummy_pipeline._data['libraryDefinitions'], str)


def test_library_definitions_in_data_str_after_calling_property_sanity(dummy_pipeline):
    assert isinstance(dummy_pipeline._library_definitions, dict)
    assert isinstance(dummy_pipeline._data['libraryDefinitions'], str)  # Check data didn't change


def test_library_definitions_lazy_loading_data_collector(dummy_pipeline, mocker):
    def get_defs_dict():
        return {"foo": "bar"}

    def get_defs_str(val):
        return '{"foo": "bar"}'

    mocker.patch('json.dumps', side_effect=get_defs_str)

    # Use mocker.Mock() here so that _control_hub.engines.get(id=self.sdc_id) in sch_models passes, MockControlHub does
    # not have an engines attribute as of writing this test.
    dummy_pipeline._control_hub = mocker.Mock()

    dummy_pipeline.executor_type = "foo"
    dummy_pipeline.sdc_id = "bar"
    dummy_pipeline._data['libraryDefinitions'] = None  # We want to trigger lazy loading in the next call

    # Lazy loading will be triggered, and we mock the json.dumps to return a string
    # We expect the property to return dict, but the underlying data to return str
    assert isinstance(dummy_pipeline._library_definitions, dict)
    assert isinstance(dummy_pipeline._data['libraryDefinitions'], str)
    assert dummy_pipeline._library_definitions == get_defs_dict()
    assert dummy_pipeline._data['libraryDefinitions'] == get_defs_str(None)


def test_library_definitions_lazy_loading_snowflake(dummy_pipeline):
    dummy_pipeline.executor_type = SNOWFLAKE_EXECUTOR_TYPE
    dummy_pipeline._data['libraryDefinitions'] = None  # We want to trigger lazy loading in the next call

    # Lazy loading will be triggered, and we mock the json.dumps to return a string
    # We expect the property to return dict, but the underlying data to return str
    assert isinstance(dummy_pipeline._library_definitions, dict)
    assert isinstance(dummy_pipeline._data['libraryDefinitions'], str)


def test_pipeline_get_stages():
    supported_connection_type = 'FOO_BAR_CONNECTION_TYPE'
    stage = {
        'instanceName': 'foo',
        'stageName': 'com_streamsets_foo_stage',
        'stageVersion': '1',
        'configuration': [{'name': 'connection', 'value': 'MANUAL'}],
        'inputLanes': [],
        'outputLanes': [],
        'uiInfo': {
            'stageType': 'SOURCE',
        },
    }
    stage_definitions = {
        'instanceName': 'foo',
        'name': 'com_streamsets_foo_stage',
        'fieldName': 'connectionSelection',
        'stageVersion': '1',
        'configuration': [{'name': 'connection', 'value': 'MANUAL'}],
        'inputLanes': [],
        'outputLanes': [],
        'configDefinitions': [
            {
                'fieldName': 'connectionSelection',
                'connectionType': supported_connection_type,
                'name': 'conf.connectionSelection',
            }
        ],
        'connectionType': 'STREAMSETS_FOO_CLIENT',
    }
    pipeline = {'pipelineId': 1, 'executorType': 'COLLECLTOR', 'name': 'foo_pipeline', 'version': 1}
    pipeline_definition = {'title': 'foo_pipeline', 'stages': [stage]}
    library_definitions = {'schemaVersion': 1, 'stages': [stage_definitions]}
    rules_definition = {}

    p = Pipeline(
        pipeline=pipeline,
        builder=None,
        pipeline_definition=pipeline_definition,
        rules_definition=rules_definition,
        library_definitions=library_definitions,
        control_hub=MockControlHub(),
    )

    stages = p.stages
    assert len(stages) == 1  # We expect one stage - see definitions above
    assert stages[0].supported_connection_types == [supported_connection_type]


def test_property_stages_not_exist_in_pipeline_definition():
    supported_connection_type = 'FOO_BAR_CONNECTION_TYPE'
    stage = {
        'instanceName': 'foo',
        'stageName': 'com_streamsets_foo_stage',
        'stageVersion': '1',
        'configuration': [{'name': 'connection', 'value': 'MANUAL'}],
        'inputLanes': [],
        'outputLanes': [],
        'uiInfo': {
            'stageType': 'SOURCE',
        },
    }
    stage_definitions = {
        'instanceName': 'foo',
        'name': 'com_streamsets_foo_stage',
        'fieldName': 'connectionSelection',
        'stageVersion': '1',
        'configuration': [{'name': 'connection', 'value': 'MANUAL'}],
        'inputLanes': [],
        'outputLanes': [],
        'configDefinitions': [
            {
                'fieldName': 'connectionSelection',
                'connectionType': supported_connection_type,
                'name': 'conf.connectionSelection',
            }
        ],
        'connectionType': 'STREAMSETS_FOO_CLIENT',
    }
    pipeline = {'pipelineId': 1, 'executorType': 'COLLECTOR', 'name': 'foo_pipeline', 'version': 1}
    pipeline_definition = {'title': 'foo_pipeline', 'stages': [stage]}
    library_definitions = {'schemaVersion': 1, 'stages': [stage_definitions]}
    rules_definition = {}

    p = Pipeline(
        pipeline=pipeline,
        builder=None,
        pipeline_definition=pipeline_definition,
        rules_definition=rules_definition,
        library_definitions=library_definitions,
        control_hub=MockControlHub(),
    )

    assert not p.error_stage
    assert not p.stats_aggregator_stage


def test_invalid_status_pipeline():
    def run_pipeline_preview():
        invalid_pipeline_preview_response = {
            'previewerId': '45e1ff73-8bd1-41be-995d-668147f1e1e2',
            'status': 'INVALID',
            'pipelineId': get_random_string,
            'attributes': {},
        }

        response = MockResponse(invalid_pipeline_preview_response, 200)
        current_status = response.json()['status']
        if current_status in STATUS_ERRORS:
            raise STATUS_ERRORS.get(current_status)(response.json())

    with pytest.raises(InvalidError) as e:
        run_pipeline_preview()

    assert e.type is InvalidError


def test_pipeline_engine_id(dummy_pipeline):
    fake_engine_id = 'abcd1234'
    assert dummy_pipeline.engine_id
    dummy_pipeline.engine_id = fake_engine_id
    assert dummy_pipeline._data['sdcId'] == fake_engine_id


def test_pipeline_sdc_id(dummy_pipeline):
    fake_engine_id = 'wxyz0987'
    assert dummy_pipeline.sdc_id
    dummy_pipeline.sdc_id = fake_engine_id
    assert dummy_pipeline._data['sdcId'] == fake_engine_id


def test_pipeline_sdc_id_is_deprecated(dummy_pipeline):
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        assert dummy_pipeline.sdc_id
        assert len(w) == 1
        assert issubclass(w[-1].category, DeprecationWarning)



# Copyright Streamsets 2023
import random

# fmt: off
import pytest

from streamsets.sdk.constants import (
    DEFAULT_MAX_CPU_LOAD_VALUE, DEFAULT_MAX_MEMORY_USED_VALUE, DEFAULT_MAX_PIPELINES_RUNNING_VALUE,
)
from streamsets.sdk.utils import get_random_string

# fmt: on

DEFAULT_LOCATION_VALUE = None
DEFAULT_LABEL_VALUE = None
DEFAULT_TAGS_VALUE = None
DEFAULT_ENGINE_BUILD_VALUE = None
DEFAULT_SCALA_BINARY_VALUE = None


class MockEnvironment:
    def __init__(self):
        self.environment_type = 'SELF'
        self.environment_id = '123456789'


@pytest.fixture(scope="function")
def passed_arguments():
    passed_arguments = {
        'deployment_name': 'deployment_name_{}'.format(get_random_string()),
        'engine_type': 'DC',
        'engine_version': '5.1.0',
        'environment': MockEnvironment(),
        'external_resource_location': 'https::{}.com'.format(get_random_string()),
        'engine_labels': ['engine_label_{}'.format(get_random_string()), 'engine_label_{}'.format(get_random_string())],
        'max_cpu_load': random.randint(1, 10),
        'max_memory_used': random.randint(1, 100),
        'max_pipelines_running': random.randint(1, 100000),
        'deployment_tags': [
            'deployment_tag_{}'.format(get_random_string()),
            'deployment_tag_{}'.format(get_random_string()),
        ],
        'engine_build': 'engine_build_{}'.format(get_random_string()),
        'scala_binary_version': '1.4.0',
    }
    return passed_arguments


@pytest.mark.parametrize(
    "deployment_builder, deployment_type",
    [
        ('self_deployment_builder', 'SELF'),
        ('gce_deployment_builder', 'GCE'),
        ('ec2_deployment_builder', 'EC2'),
        ('azure_deployment_builder', 'AZURE_VM'),
        ('kubernetes_deployment_builder', 'KUBERNETES'),
    ],
)
def test_deployment_builder_build_sanity(deployment_builder, deployment_type, passed_arguments, request):
    deployment_builder = request.getfixturevalue(deployment_builder)

    passed_arguments['environment'].environment_type = deployment_type

    deployment = deployment_builder.build(
        deployment_name=passed_arguments['deployment_name'],
        environment=passed_arguments['environment'],
        engine_type=passed_arguments['engine_type'],
        engine_version=passed_arguments['engine_version'],
    )
    assert deployment


@pytest.mark.parametrize(
    "deployment_builder, deployment_type",
    [
        ('self_deployment_builder', 'SELF'),
        ('gce_deployment_builder', 'GCE'),
        ('ec2_deployment_builder', 'EC2'),
        ('azure_deployment_builder', 'AZURE_VM'),
        ('kubernetes_deployment_builder', 'KUBERNETES'),
    ],
)
def test_deployment_builder_build_respects_arguments_passed(
    deployment_builder, passed_arguments, deployment_type, request
):
    passed_arguments['environment'].environment_type = deployment_type

    deployment_builder = request.getfixturevalue(deployment_builder)

    deployment = deployment_builder.build(
        deployment_name=passed_arguments['deployment_name'],
        engine_type=passed_arguments['engine_type'],
        engine_version=passed_arguments['engine_version'],
        environment=passed_arguments['environment'],
        external_resource_location=passed_arguments['external_resource_location'],
        engine_labels=passed_arguments['engine_labels'],
        max_cpu_load=passed_arguments['max_cpu_load'],
        max_memory_used=passed_arguments['max_memory_used'],
        max_pipelines_running=passed_arguments['max_pipelines_running'],
        deployment_tags=passed_arguments['deployment_tags'],
        engine_build=passed_arguments['engine_build'],
        scala_binary_version=passed_arguments['scala_binary_version'],
    )

    # Make sure all passed arguments are respected
    assert deployment._data['name'] == passed_arguments['deployment_name']
    assert deployment._data['type'] == deployment_type
    assert deployment._data['engineVersion'] == passed_arguments['engine_version']
    assert deployment._data['envId'] == passed_arguments['environment'].environment_id
    assert (
        deployment._data['engineConfiguration']['externalResourcesUri']
        == passed_arguments['external_resource_location']
    )
    assert deployment._data['engineConfiguration']['labels'] == passed_arguments['engine_labels']
    assert deployment._data['engineConfiguration']['maxCpuLoad'] == passed_arguments['max_cpu_load']
    assert deployment._data['engineConfiguration']['maxMemoryUsed'] == passed_arguments['max_memory_used']
    assert deployment._data['engineConfiguration']['maxPipelinesRunning'] == passed_arguments['max_pipelines_running']
    assert [deployment['tag'] for deployment in deployment._data['deploymentTags']] == passed_arguments[
        'deployment_tags'
    ]
    assert deployment._data['engineBuild'] == passed_arguments['engine_build']
    assert deployment._data['scalaBinaryVersion'] == passed_arguments['scala_binary_version']

    if deployment_type == "AWS":
        assert deployment._data['sshKeySource'] == "NONE"


@pytest.mark.parametrize(
    "deployment_builder, deployment_type",
    [
        ('self_deployment_builder', 'SELF'),
        ('gce_deployment_builder', 'GCE'),
        ('ec2_deployment_builder', 'EC2'),
        ('azure_deployment_builder', 'AZURE_VM'),
        ('kubernetes_deployment_builder', 'KUBERNETES'),
    ],
)
def test_deployment_builder_build_respects_default_values(
    deployment_builder, deployment_type, passed_arguments, request
):
    deployment_builder = request.getfixturevalue(deployment_builder)
    passed_arguments['environment'].environment_type = deployment_type

    deployment = deployment_builder.build(
        deployment_name=passed_arguments['deployment_name'],
        environment=passed_arguments['environment'],
        engine_type=passed_arguments['engine_type'],
        engine_version=passed_arguments['engine_version'],
    )

    # Make sure default values are respected
    assert deployment._data['engineConfiguration']['externalResourcesUri'] == DEFAULT_LOCATION_VALUE
    assert deployment._data['engineConfiguration']['labels'] == DEFAULT_LABEL_VALUE
    assert deployment._data['engineConfiguration']['maxCpuLoad'] == DEFAULT_MAX_CPU_LOAD_VALUE
    assert deployment._data['engineConfiguration']['maxMemoryUsed'] == DEFAULT_MAX_MEMORY_USED_VALUE
    assert deployment._data['engineConfiguration']['maxPipelinesRunning'] == DEFAULT_MAX_PIPELINES_RUNNING_VALUE
    assert deployment._data['deploymentTags'] == DEFAULT_TAGS_VALUE
    assert deployment._data['engineBuild'] == DEFAULT_ENGINE_BUILD_VALUE
    assert deployment._data['scalaBinaryVersion'] == DEFAULT_SCALA_BINARY_VALUE


@pytest.mark.parametrize(
    "key_to_change, value",
    [
        ('deployment_name', 1),
        ('engine_type', 1),
        ('engine_version', 1),
        ('external_resource_location', 1),
        ('engine_labels', 'STRING VALUE'),
        ('max_cpu_load', 'STRING VALUE'),
        ('max_memory_used', 'STRING VALUE'),
        ('max_pipelines_running', 'STRING VALUE'),
        ('deployment_tags', 'STRING VALUE'),
        ('engine_build', 1),
        ('scala_binary_version', 1),
    ],
)
def test_deployment_builder_build_fails_on_invalid_types(
    key_to_change, value, self_deployment_builder, passed_arguments, request
):
    passed_arguments['environment'].environment_type = 'SELF'
    passed_arguments[key_to_change] = value

    with pytest.raises(TypeError):
        self_deployment_builder.build(
            deployment_name=passed_arguments['deployment_name'],
            engine_type=passed_arguments['engine_type'],
            engine_version=passed_arguments['engine_version'],
            environment=passed_arguments['environment'],
            external_resource_location=passed_arguments['external_resource_location'],
            engine_labels=passed_arguments['engine_labels'],
            max_cpu_load=passed_arguments['max_cpu_load'],
            max_memory_used=passed_arguments['max_memory_used'],
            max_pipelines_running=passed_arguments['max_pipelines_running'],
            deployment_tags=passed_arguments['deployment_tags'],
            engine_build=passed_arguments['engine_build'],
            scala_binary_version=passed_arguments['scala_binary_version'],
        )


@pytest.mark.parametrize(
    "deployment_builder, deployment_type",
    [
        ('self_deployment_builder', 'SELF'),
        ('gce_deployment_builder', 'GCE'),
        ('ec2_deployment_builder', 'EC2'),
        ('azure_deployment_builder', 'AZURE_VM'),
        ('kubernetes_deployment_builder', 'KUBERNETES'),
    ],
)
def test_deployment_builder_build_fails_for_value_over_100_for_max_cpu_load(
    deployment_builder, passed_arguments, deployment_type, request
):
    passed_arguments['environment'].environment_type = deployment_type

    deployment_builder = request.getfixturevalue(deployment_builder)

    with pytest.raises(ValueError):
        deployment_builder.build(
            deployment_name=passed_arguments['deployment_name'],
            engine_type=passed_arguments['engine_type'],
            engine_version=passed_arguments['engine_version'],
            environment=passed_arguments['environment'],
            external_resource_location=passed_arguments['external_resource_location'],
            engine_labels=passed_arguments['engine_labels'],
            max_cpu_load=1000.0,
        )


@pytest.mark.parametrize(
    "deployment_builder, deployment_type",
    [
        ('self_deployment_builder', 'SELF'),
        ('gce_deployment_builder', 'GCE'),
        ('ec2_deployment_builder', 'EC2'),
        ('azure_deployment_builder', 'AZURE_VM'),
        ('kubernetes_deployment_builder', 'KUBERNETES'),
    ],
)
def test_deployment_builder_build_fails_for_value_over_100_for_max_memory_used(
    deployment_builder, passed_arguments, deployment_type, request
):
    passed_arguments['environment'].environment_type = deployment_type

    deployment_builder = request.getfixturevalue(deployment_builder)

    with pytest.raises(ValueError):
        deployment_builder.build(
            deployment_name=passed_arguments['deployment_name'],
            engine_type=passed_arguments['engine_type'],
            engine_version=passed_arguments['engine_version'],
            environment=passed_arguments['environment'],
            external_resource_location=passed_arguments['external_resource_location'],
            engine_labels=passed_arguments['engine_labels'],
            max_memory_used=1000.0,
        )
